import asyncio
import base64
from collections.abc import Iterable
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Any, Dict, List, Optional, TypedDict, Union
from uuid import uuid4

import playwright.sync_api
from playwright.sync_api import Page
from playwright.sync_api import TimeoutError as PlaywrightTimeoutError

from .base import AgentTools


class ClientCertificate(TypedDict):
    origin: str
    pfx_base64: str
    passphrase: str


class ElementExtractor:
    def __init__(self):
        self.interactive_selectors = [
            "a[href]",
            "button",
            "input",
            "textarea",
            "select",
            "input[type='button']",
            "input[type='submit']",
            "[role='button']",
            "[onclick]",
            "[tabindex]:not([tabindex='-1'])",
        ]

    def extract_elements(self, page: Page) -> List[Dict[str, Any]]:
        elements = []

        js_code = """
        () => {
            const elements = [];
            const selectors = [
                'a[href]',
                'button',
                'input',
                'textarea',
                'select',
                '[role="button"]',
                '[onclick]',
                '[tabindex]:not([tabindex="-1"])'
            ];

            const seen = new Set();

            function getLabel(el) {
                const ariaLabel = el.getAttribute('aria-label');
                if (ariaLabel) return ariaLabel;

                const title = el.getAttribute('title');
                if (title) return title;

                if (el.id) {
                    const label = document.querySelector(`label[for="${CSS.escape(el.id)}"]`);
                    if (label) return label.innerText.trim();
                }

                const parentLabel = el.closest('label');
                if (parentLabel) return parentLabel.innerText.trim();

                const placeholder = el.getAttribute('placeholder');
                if (placeholder) return `[${placeholder}]`;

                return '';
            }

            function getParentContext(el) {
                const semanticParents = [
                    'form', 'nav', 'header', 'footer', 'main', 'aside', 'section', 'article', 'dialog',
                    '[role="dialog"]', '[role="navigation"]', '[role="form"]', '[role="banner"]', '[role="contentinfo"]'
                ];

                const parent = el.closest(semanticParents.join(','));
                if (!parent) return 'BODY';

                let context = parent.tagName.toLowerCase();

                if (parent.getAttribute('aria-label')) context += ` (${parent.getAttribute('aria-label')})`;
                else if (parent.id) context += ` (#${parent.id})`;
                else if (parent.getAttribute('role')) context += ` [role="${parent.getAttribute('role')}"]`;

                return context.toUpperCase();
            }

            selectors.forEach(selector => {
                document.querySelectorAll(selector).forEach((el, idx) => {
                    if (seen.has(el)) return;
                    seen.add(el);

                    const rect = el.getBoundingClientRect();
                    const style = window.getComputedStyle(el);

                    const isVisible = rect.width > 0 &&
                                    rect.height > 0 &&
                                    style.visibility !== 'hidden' &&
                                    style.display !== 'none' &&
                                    style.opacity !== '0';

                    if (!isVisible) return;

                    const vw = window.innerWidth;
                    const vh = window.innerHeight;
                    const isOnScreen = rect.right > 0 &&
                                       rect.bottom > 0 &&
                                       rect.left < vw &&
                                       rect.top < vh;

                    // Check if element is actually the topmost at its center.
                    // Filters elements hidden behind sticky headers, overlays,
                    // or positioned at (0,0) behind page chrome.
                    let isTopmost = true;
                    if (isOnScreen) {
                        const cx = rect.left + rect.width / 2;
                        const cy = rect.top + rect.height / 2;
                        if (cx >= 0 && cy >= 0 && cx < vw && cy < vh) {
                            const topEl = document.elementFromPoint(cx, cy);
                            isTopmost = !topEl || topEl === el || el.contains(topEl);
                        }
                    }

                    let uniqueSelector = '';
                    if (el.id) {
                        uniqueSelector = `#${CSS.escape(el.id)}`;
                    } else {
                        let path = [];
                        let current = el;
                        while (current && current !== document.body) {
                            let selector = current.tagName.toLowerCase();

                            if (current.className && typeof current.className === 'string') {
                                const classes = current.className.split(' ')
                                    .filter(c => c && !c.startsWith('css-'))
                                    .slice(0, 2);
                                if (classes.length > 0) {
                                    selector += '.' + classes.join('.');
                                }
                            }

                            if (current.parentElement) {
                                const siblings = Array.from(current.parentElement.children);
                                const sameTagSiblings = siblings.filter(s => s.tagName === current.tagName);
                                if (sameTagSiblings.length > 1) {
                                    const index = sameTagSiblings.indexOf(current) + 1;
                                    selector += `:nth-of-type(${index})`;
                                }
                            }

                            path.unshift(selector);
                            current = current.parentElement;
                        }
                        uniqueSelector = path.join(' > ');
                    }

                    const label = getLabel(el);
                    const parentContext = getParentContext(el);

                    let text = el.innerText?.trim().substring(0, 100) ||
                               el.getAttribute('aria-label') ||
                               el.getAttribute('title') ||
                               el.value ||
                               el.alt ||
                               label ||
                               '';

                    const pageX = rect.x + window.scrollX;
                    const pageY = rect.y + window.scrollY;

                    let category;
                    const isInDropdown = el.closest('.dropdown-menu, [role="listbox"], [role="menu"]');
                    if (isInDropdown) {
                        category = 'dropdown-option';
                    } else if (el.tagName === 'BUTTON' || el.type === 'submit' || el.type === 'button' || el.getAttribute('role') === 'button') {
                        category = 'action-button';
                    } else if (el.tagName === 'INPUT' || el.tagName === 'TEXTAREA' || el.tagName === 'SELECT') {
                        category = 'form-input';
                    } else if (el.tagName === 'A' && el.href) {
                        category = 'navigation-link';
                    } else {
                        category = 'other';
                    }

                    elements.push({
                        selector: uniqueSelector,
                        text: text,
                        tag: el.tagName.toLowerCase(),
                        category: category,
                        parent_context: parentContext,
                        attributes: {
                            href: el.href || '',
                            onclick: el.onclick ? 'yes' : '',
                            role: el.getAttribute('role') || '',
                            type: el.type || '',
                            tabindex: el.tabIndex || 0,
                            placeholder: el.getAttribute('placeholder') || '',
                            name: el.name || '',
                            ariaLabel: el.getAttribute('aria-label') || '',
                            title: el.getAttribute('title') || '',
                            dataTestId: el.getAttribute('data-testid') || '',
                            label: label
                        },
                        bbox: {
                            x: Math.round(pageX),
                            y: Math.round(pageY),
                            width: Math.round(rect.width),
                            height: Math.round(rect.height)
                        },
                        viewport_bbox: {
                            x: Math.round(rect.x),
                            y: Math.round(rect.y),
                            width: Math.round(rect.width),
                            height: Math.round(rect.height)
                        },
                        isVisible: rect.width > 0 && rect.height > 0,
                        isOnScreen: isOnScreen,
                        isTopmost: isTopmost
                    });
                });
            });

            const visible = elements.filter(e => e.isVisible && e.isOnScreen && e.isTopmost);

            // Keep DOM order but move dropdown options to the end.
            // Dropdown items (e.g. 88 currency options) bury critical
            // elements like submit buttons when listed inline.
            const main = visible.filter(e => e.category !== 'dropdown-option');
            const dropdown = visible.filter(e => e.category === 'dropdown-option');
            return main.concat(dropdown);
        }
        """

        try:
            raw_elements = page.evaluate(js_code)

            for idx, elem in enumerate(raw_elements):
                elem["index"] = idx
                elements.append(elem)

        except Exception as e:
            print(f"[WARN][ElementExtractor] Error extracting elements: {e}")

        return elements

    def get_element_by_index(
        self, elements: List[Dict[str, Any]], index: int
    ) -> Optional[Dict[str, Any]]:
        for elem in elements:
            if elem["index"] == index:
                return elem
        return None


def _slim_element(element: Dict[str, Any]) -> Dict[str, Any]:
    return {
        "index": element["index"],
        "selector": element["selector"],
        "text": element["text"],
        "tag": element["tag"],
        "category": element["category"],
        "parent_context": element["parent_context"],
        "type": element.get("attributes", {}).get("type", ""),
        "href": element.get("attributes", {}).get("href", ""),
    }


def to_urls(u: Optional[Union[str, Iterable[str]]] = None) -> Optional[Iterable[str]]:
    if u is None:
        return None
    if isinstance(u, str):
        return [u]
    return u


def _prepare_script(script: str) -> str:
    """Wrap bare `return` statements in an async IIFE for Playwright evaluate()."""
    if "return " in script and not script.strip().startswith("("):
        return f"(async () => {{ {script} }})()"
    return script


def _is_target_closed(e: Exception) -> bool:
    msg = str(e)
    return "Target" in msg and "closed" in msg


class BrowserTools(AgentTools):
    urls: Optional[Iterable[str]]
    browser: playwright.sync_api.Browser
    pages: Dict[str, playwright.sync_api.Page]
    listen_network: bool
    listen_console: bool
    network_requests: Dict[str, List[playwright.sync_api.Request]]
    console_logs: Dict[str, List[playwright.sync_api.ConsoleMessage]]
    debug_mode: bool
    allow_close_page: bool
    headless: bool

    def __init__(
        self,
        url: Optional[Union[str, Iterable[str]]] = None,
        listen_network: bool = False,
        listen_console: bool = False,
        debug_mode: bool = False,
        allow_close_page: bool = True,
        headless: bool = True,
        client_certificate: Optional[ClientCertificate] = None,
    ):
        self.urls = to_urls(url)
        self.debug_mode = debug_mode
        self.allow_close_page = allow_close_page
        self.headless = headless
        if self.debug_mode:
            print(
                f"[DEBUG][BrowserTools.__init__] url={url}, listen_network={listen_network}, listen_console={listen_console}, debug_mode={debug_mode}, allow_close_page={allow_close_page}"
            )
            print(f"[DEBUG][BrowserTools.__init__] resolved urls={self.urls}")
        try:
            loop = asyncio.get_running_loop()
            if loop.is_running():
                if self.debug_mode:
                    print(
                        "[DEBUG][BrowserTools.__init__] Found running event loop, unsetting it"
                    )
                asyncio._set_running_loop(None)
        except RuntimeError:
            if self.debug_mode:
                print("[DEBUG][BrowserTools.__init__] No running event loop found")
        self._playwright_context = playwright.sync_api.sync_playwright()
        pw = self._playwright_context.start()
        if self.debug_mode:
            print(
                f"[DEBUG][BrowserTools.__init__] Playwright started, launching chromium (headless={self.headless})"
            )
        self.browser = pw.chromium.launch(headless=self.headless)
        if self.debug_mode:
            print("[DEBUG][BrowserTools.__init__] Browser launched successfully")
        self._browser_context = self._build_browser_context(client_certificate)
        self.pages = {}
        self.listen_network = listen_network
        self.listen_console = listen_console
        self.network_requests = {}
        self.console_logs = {}
        self._extracted_elements: Dict[str, List[Dict[str, Any]]] = {}
        self.extractor = ElementExtractor()

    def _build_browser_context(
        self, client_certificate: Optional[ClientCertificate]
    ) -> playwright.sync_api.BrowserContext:
        context_options: Dict[str, Any] = {}

        if client_certificate is not None:
            origin: str = client_certificate["origin"]
            pfx_base64: str = client_certificate["pfx_base64"]
            passphrase: str = client_certificate["passphrase"]

            if self.debug_mode:
                print(
                    f"[DEBUG][BrowserTools._build_browser_context] Using client certificate for origin={origin}"
                )

            context_options["client_certificates"] = [
                {
                    "origin": origin,
                    "pfx": base64.b64decode(pfx_base64),
                    "passphrase": passphrase,
                }
            ]
            context_options["ignore_https_errors"] = True

        return self.browser.new_context(**context_options)

    def close(self):
        if self.debug_mode:
            print(f"[DEBUG][BrowserTools.close] Closing {len(self.pages)} pages")
        for page_id, page in list(self.pages.items()):
            try:
                if self.debug_mode:
                    print(f"[DEBUG][BrowserTools.close] Closing page {page_id}")
                page.close()
            except Exception as e:
                if self.debug_mode:
                    print(
                        f"[DEBUG][BrowserTools.close] Error closing page {page_id}: {e}"
                    )
        self.pages.clear()

        try:
            if self.debug_mode:
                print("[DEBUG][BrowserTools.close] Closing browser context")
            self._browser_context.close()
        except Exception as e:
            if self.debug_mode:
                print(f"[DEBUG][BrowserTools.close] Error closing browser context: {e}")

        try:
            if self.debug_mode:
                print("[DEBUG][BrowserTools.close] Closing browser")
            self.browser.close()
        except Exception as e:
            if self.debug_mode:
                print(f"[DEBUG][BrowserTools.close] Error closing browser: {e}")

        try:
            if self.debug_mode:
                print("[DEBUG][BrowserTools.close] Exiting playwright context")
            self._playwright_context.__exit__(None, None, None)
        except Exception as e:
            if self.debug_mode:
                print(
                    f"[DEBUG][BrowserTools.close] Error exiting playwright context: {e}"
                )

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    def _get_page(self, page_id: str) -> Page:
        if page_id not in self.pages:
            raise ValueError(f"Page '{page_id}' does not exist.")
        return self.pages[page_id]

    def _handle_page_crash(self, page_id: str):
        self.pages.pop(page_id, None)
        self._extracted_elements.pop(page_id, None)
        raise RuntimeError(
            "Browser page has closed or crashed. "
            "Call navigate_to_url to start a new session."
        )

    def _get_page_id_by_request(
        self, request: playwright.sync_api.Request
    ) -> Optional[str]:
        for page_id, page in self.pages.items():
            if request.frame.page == page:
                return page_id
        return None

    def _get_page_id_by_console_message(
        self, msg: playwright.sync_api.ConsoleMessage
    ) -> Optional[str]:
        for page_id, page in self.pages.items():
            if msg.page == page:
                return page_id
        return None

    def _attach_listeners(self, page: playwright.sync_api.Page):
        if self.listen_network:

            def on_request(request: playwright.sync_api.Request):
                page_id = self._get_page_id_by_request(request)
                if page_id:
                    if page_id not in self.network_requests:
                        self.network_requests[page_id] = []
                    self.network_requests[page_id].append(request)

            page.on("request", on_request)

        if self.listen_console:

            def on_console_message(msg: playwright.sync_api.ConsoleMessage):
                page_id = self._get_page_id_by_console_message(msg)
                if page_id:
                    if page_id not in self.console_logs:
                        self.console_logs[page_id] = []
                    self.console_logs[page_id].append(msg)

            page.on("console", on_console_message)

    def navigate_to_url(
        self, url: str, page_id: Optional[str] = None
    ) -> Dict[str, str]:
        """Navigate to a URL. If page_id is provided, reuses that page; otherwise creates a new page. Returns page_id, final url, and title. After navigation, call get_page_summary to discover interactive elements on the new page."""
        if self.debug_mode:
            print(f"[DEBUG][BrowserTools.navigate_to_url] url={url}, page_id={page_id}")
        if self.urls is not None and url not in self.urls:
            if self.debug_mode:
                print(
                    f"[DEBUG][BrowserTools.navigate_to_url] URL not allowed. allowed_urls={list(self.urls)}"
                )
            raise PermissionError(f"URL '{url}' is not allowed.")

        if page_id is None:
            page_id = str(uuid4())
            if self.debug_mode:
                print(
                    f"[DEBUG][BrowserTools.navigate_to_url] Creating new page with id={page_id}"
                )
            try:
                page = self._browser_context.new_page()
            except Exception as e:
                if _is_target_closed(e):
                    raise RuntimeError(
                        "Browser has closed. Cannot create new pages. "
                        "The browser session may have expired."
                    )
                raise
            if page is None:
                raise RuntimeError("Failed to create a new browser page.")
            self.pages[page_id] = page
            self._attach_listeners(page)
        else:
            page = self._get_page(page_id)
            if self.debug_mode:
                print(
                    f"[DEBUG][BrowserTools.navigate_to_url] Reusing existing page {page_id}"
                )

        if self.debug_mode:
            print(f"[DEBUG][BrowserTools.navigate_to_url] Navigating to {url}")

        try:
            page.goto(url)
        except Exception as e:
            if _is_target_closed(e):
                self._handle_page_crash(page_id)
            raise

        self._extracted_elements.pop(page_id, None)
        if self.debug_mode:
            print(
                f"[DEBUG][BrowserTools.navigate_to_url] Navigation complete. page.url={page.url}, title={page.title()}"
            )
        return dict(page_id=page_id, url=page.url, title=page.title())

    def click(
        self,
        page_id: str,
        selector: Optional[str] = None,
        x: Optional[float] = None,
        y: Optional[float] = None,
    ):
        """Click an element by CSS selector, or click at specific x,y coordinates. Provide either selector OR both x and y. IMPORTANT: Use selectors from get_page_summary or get_element_by_summary_index. Prefer click_element(page_id, index) to avoid selector errors."""
        if self.debug_mode:
            print(
                f"[DEBUG][BrowserTools.click] page_id={page_id}, selector={selector}, x={x}, y={y}"
            )
        page = self._get_page(page_id)

        try:
            if selector is not None:
                element = page.query_selector(selector)
                if element is None:
                    available = self._get_available_selectors_hint(page_id)
                    raise ValueError(
                        f"Selector '{selector}' not found on the page. "
                        f"Use click_element(page_id, index) with an index from get_page_summary instead. "
                        f"{available}"
                    )
                page.click(selector, timeout=5000)
            elif x is not None and y is not None:
                page.mouse.click(x, y)
            else:
                raise ValueError(
                    "Provide either 'selector' OR both 'x' and 'y' coordinates."
                )
        except ValueError:
            raise
        except PlaywrightTimeoutError:
            available = self._get_available_selectors_hint(page_id)
            raise ValueError(
                f"Element '{selector}' found but not clickable (timeout). "
                f"Use click_element(page_id, index) with an index from get_page_summary. "
                f"{available}"
            )
        except Exception as e:
            if _is_target_closed(e):
                self._handle_page_crash(page_id)
            raise

        if self.debug_mode:
            print(f"[DEBUG][BrowserTools.click] Click complete. Current url={page.url}")

    def click_element(self, page_id: str, index: int):
        """Click an interactive element by its index from the last get_page_summary. This is the preferred way to click — pass the index number directly instead of a CSS selector. Call get_page_summary first to see available elements and their indices."""
        if self.debug_mode:
            print(
                f"[DEBUG][BrowserTools.click_element] page_id={page_id}, index={index}"
            )
        element = self._resolve_element(page_id, index)
        self.click(page_id, selector=element["selector"])
        return element

    def _resolve_element(self, page_id: str, index: int) -> Dict[str, Any]:
        page = self._get_page(page_id)
        cached = self._extracted_elements.get(page_id)
        if cached is None:
            cached = self.extractor.extract_elements(page)
            self._extracted_elements[page_id] = cached

        element = self.extractor.get_element_by_index(cached, index)

        if element is not None:
            selector = element.get("selector", "")
            try:
                if selector and not page.query_selector(selector):
                    if self.debug_mode:
                        print(
                            f"[DEBUG][BrowserTools._resolve_element] Stale selector at index {index}, re-extracting"
                        )
                    cached = self.extractor.extract_elements(page)
                    self._extracted_elements[page_id] = cached
                    element = self.extractor.get_element_by_index(cached, index)
            except Exception as e:
                if _is_target_closed(e):
                    self._handle_page_crash(page_id)
                raise

        if element is None:
            count = len(cached)
            raise ValueError(
                f"No element at index {index}. "
                f"Page has {count} elements (indices 0-{count - 1}). "
                f"Call get_page_summary to see current elements."
            )
        return element

    def _get_available_selectors_hint(self, page_id: str) -> str:
        cached = self._extracted_elements.get(page_id)
        if cached is None:
            return ""
        selectors = [
            e["selector"]
            for e in cached
            if e.get("selector") and not e["selector"].startswith("form.")
        ]
        if not selectors:
            return ""
        preview = selectors[:10]
        hint = "Available selectors from last page summary: " + ", ".join(preview)
        if len(selectors) > 10:
            hint += f" ... and {len(selectors) - 10} more"
        return hint

    def fill(self, page_id: str, selector: str, value: str):
        """Fill a form field with a value using a CSS selector. IMPORTANT: Use selectors from get_page_summary or get_element_by_summary_index. Prefer fill_element(page_id, index, value) to avoid selector errors."""
        if self.debug_mode:
            print(
                f"[DEBUG][BrowserTools.fill] page_id={page_id}, selector={selector}, value={value!r}"
            )
        page = self._get_page(page_id)

        try:
            element = page.query_selector(selector)
            if element is None:
                available = self._get_available_selectors_hint(page_id)
                raise ValueError(
                    f"Selector '{selector}' not found on the page. "
                    f"Use fill_element(page_id, index, value) with an index from get_page_summary instead. "
                    f"{available}"
                )
            page.fill(selector, value, timeout=5000)
        except ValueError:
            raise
        except PlaywrightTimeoutError:
            available = self._get_available_selectors_hint(page_id)
            raise ValueError(
                f"Element '{selector}' found but not fillable (timeout). "
                f"Use fill_element(page_id, index, value) with an index from get_page_summary. "
                f"{available}"
            )
        except Exception as e:
            if _is_target_closed(e):
                self._handle_page_crash(page_id)
            raise

        if self.debug_mode:
            print("[DEBUG][BrowserTools.fill] Fill complete")

    def fill_element(self, page_id: str, index: int, value: str):
        """Fill a form field by its index from the last get_page_summary. This is the preferred way to fill — pass the index number directly instead of a CSS selector. Call get_page_summary first to see available elements and their indices."""
        if self.debug_mode:
            print(
                f"[DEBUG][BrowserTools.fill_element] page_id={page_id}, index={index}, value={value!r}"
            )
        element = self._resolve_element(page_id, index)
        self.fill(page_id, selector=element["selector"], value=value)
        return element

    def get_html(self, page_id: str) -> str:
        """Get the full HTML content of the page. Prefer get_page_summary for identifying interactive elements — use this only when you need raw HTML inspection."""
        if self.debug_mode:
            print(f"[DEBUG][BrowserTools.get_html] page_id={page_id}")
        page = self._get_page(page_id)
        try:
            content = page.content()
        except Exception as e:
            if _is_target_closed(e):
                self._handle_page_crash(page_id)
            raise
        if self.debug_mode:
            print(
                f"[DEBUG][BrowserTools.get_html] Got HTML content, length={len(content)}"
            )
        return content

    def get_text(self, page_id: str, selector: str) -> str:
        """Get the inner text content of an element by CSS selector."""
        if self.debug_mode:
            print(
                f"[DEBUG][BrowserTools.get_text] page_id={page_id}, selector={selector}"
            )
        page = self._get_page(page_id)
        try:
            element = page.query_selector(selector)
            if element is None:
                raise ValueError(
                    f"Selector '{selector}' not found. "
                    f"Call get_page_summary to see current elements."
                )
            text = page.inner_text(selector)
        except ValueError:
            raise
        except Exception as e:
            if _is_target_closed(e):
                self._handle_page_crash(page_id)
            raise
        if self.debug_mode:
            print(
                f"[DEBUG][BrowserTools.get_text] Got text, length={len(text)}, preview={text[:200]!r}"
            )
        return text

    def get_attribute(
        self, page_id: str, selector: str, attribute: str
    ) -> Optional[str]:
        """Get a single HTML attribute value from an element by CSS selector."""
        if self.debug_mode:
            print(
                f"[DEBUG][BrowserTools.get_attribute] page_id={page_id}, selector={selector}, attribute={attribute}"
            )
        page = self._get_page(page_id)
        try:
            element = page.query_selector(selector)
            if element is None:
                raise ValueError(
                    f"Selector '{selector}' not found. "
                    f"Call get_page_summary to see current elements."
                )
            value = page.get_attribute(selector, attribute)
        except ValueError:
            raise
        except Exception as e:
            if _is_target_closed(e):
                self._handle_page_crash(page_id)
            raise
        if self.debug_mode:
            print(f"[DEBUG][BrowserTools.get_attribute] Result: {value!r}")
        return value

    def get_attributes(self, page_id: str, selector: str) -> Dict[str, Optional[str]]:
        """Get all HTML attributes of an element by CSS selector. Returns a dict of attribute name to value."""
        if self.debug_mode:
            print(
                f"[DEBUG][BrowserTools.get_attributes] page_id={page_id}, selector={selector}"
            )
        page = self._get_page(page_id)
        try:
            element_handle = page.query_selector(selector)
        except Exception as e:
            if _is_target_closed(e):
                self._handle_page_crash(page_id)
            raise
        if element_handle is None:
            if self.debug_mode:
                print(
                    f"[DEBUG][BrowserTools.get_attributes] No element found for selector={selector}"
                )
            raise ValueError(f"Selector '{selector}' did not match any elements.")
        attrs = element_handle.evaluate(
            "el => { const attrs = {}; for (let attr of el.attributes) { attrs[attr.name] = attr.value; } return attrs; }"
        )
        if self.debug_mode:
            print(
                f"[DEBUG][BrowserTools.get_attributes] Found {len(attrs)} attributes: {list(attrs.keys())}"
            )
        return attrs

    def get_all_links(self, page_id: str) -> List[Dict[str, str]]:
        """Get all links on the page as a list of {text, href} objects."""
        if self.debug_mode:
            print(f"[DEBUG][BrowserTools.get_all_links] page_id={page_id}")
        page = self._get_page(page_id)
        try:
            links = page.query_selector_all("a")
        except Exception as e:
            if _is_target_closed(e):
                self._handle_page_crash(page_id)
            raise
        if self.debug_mode:
            print(
                f"[DEBUG][BrowserTools.get_all_links] Found {len(links)} <a> elements"
            )
        result = []
        for link in links:
            href = link.get_attribute("href")
            if href is not None:
                result.append({"text": link.inner_text(), "href": href})
        if self.debug_mode:
            print(
                f"[DEBUG][BrowserTools.get_all_links] Returning {len(result)} links with href"
            )
        return result

    def get_network_requests(self, page_id: str) -> Iterable[dict]:
        """Get all captured network requests for a page. Requires listen_network=True on initialization."""
        if self.debug_mode:
            print(f"[DEBUG][BrowserTools.get_network_requests] page_id={page_id}")
        self._get_page(page_id)

        requests = self.network_requests.get(page_id, [])
        if self.debug_mode:
            print(
                f"[DEBUG][BrowserTools.get_network_requests] Found {len(requests)} requests for page {page_id}"
            )
            for r in requests:
                print(
                    f"[DEBUG][BrowserTools.get_network_requests]   {r.method} {r.url}"
                )

        result = []
        for request in requests:
            res = request.response()
            result.append(
                {
                    "request": {
                        "url": request.url,
                        "method": request.method,
                        "headers": dict(request.headers),
                        "post_data": request.post_data,
                    },
                    "response": {
                        "status": res.status if res else None,
                        "headers": dict(res.headers) if res else None,
                        "body": res.text() if res else None,
                    },
                }
            )
        return result

    def get_console_logs(self, page_id: str) -> Iterable[dict]:
        """Get all captured console log messages for a page. Requires listen_console=True on initialization."""
        if self.debug_mode:
            print(f"[DEBUG][BrowserTools.get_console_logs] page_id={page_id}")
        self._get_page(page_id)

        logs = self.console_logs.get(page_id, [])
        if self.debug_mode:
            print(
                f"[DEBUG][BrowserTools.get_console_logs] Found {len(logs)} console messages for page {page_id}"
            )

        return [
            {
                "type": msg.type,
                "text": msg.text,
                "location": {
                    "url": msg.location["url"],
                    "lineNumber": msg.location["lineNumber"],
                    "columnNumber": msg.location["columnNumber"],
                }
                if msg.location
                else None,
            }
            for msg in logs
        ]

    def list_pages(self) -> Iterable[dict]:
        """List all open browser pages with their page_id, URL, and title."""
        if self.debug_mode:
            print(f"[DEBUG][BrowserTools.list_pages] Total pages: {len(self.pages)}")
            for pid, p in self.pages.items():
                print(
                    f"[DEBUG][BrowserTools.list_pages]   page_id={pid}, url={p.url}, title={p.title()}"
                )
        return [
            {
                "page_id": page_id,
                "url": page.url,
                "title": page.title(),
            }
            for page_id, page in self.pages.items()
        ]

    def wait(self, page_id: str, milliseconds: int = 1000):
        """Wait for a specified number of milliseconds (default 1000). Use this instead of execute_javascript with setTimeout — it does NOT invalidate the element cache. Useful for waiting after clicks, form submissions, or page transitions before taking a screenshot or calling get_page_summary."""
        if milliseconds < 0 or milliseconds > 30000:
            raise ValueError("milliseconds must be between 0 and 30000.")
        if self.debug_mode:
            print(
                f"[DEBUG][BrowserTools.wait] page_id={page_id}, milliseconds={milliseconds}"
            )
        page = self._get_page(page_id)
        try:
            page.wait_for_timeout(milliseconds)
        except Exception as e:
            if _is_target_closed(e):
                self._handle_page_crash(page_id)
            raise
        if self.debug_mode:
            print("[DEBUG][BrowserTools.wait] Wait complete")

    def close_page(self, page_id: str):
        """Close a browser page by its page_id."""
        if self.debug_mode:
            print(f"[DEBUG][BrowserTools.close_page] page_id={page_id}")
        if page_id in self.pages:
            page = self.pages[page_id]
            page.close()
            del self.pages[page_id]
            self._extracted_elements.pop(page_id, None)
            if self.debug_mode:
                print(
                    f"[DEBUG][BrowserTools.close_page] Page closed. Remaining pages: {list(self.pages.keys())}"
                )
        else:
            raise ValueError(f"Page '{page_id}' does not exist.")

    def execute_javascript(self, page_id: str, script: str):
        """Execute JavaScript code on the page and return the result. WARNING: JavaScript execution may change the DOM. After calling this, the cached page summary is invalidated — call get_page_summary before using any selectors from a previous summary."""
        if self.debug_mode:
            print(
                f"[DEBUG][BrowserTools.execute_javascript] page_id={page_id}, script={script[:200]!r}"
            )
        page = self._get_page(page_id)
        script = _prepare_script(script)
        try:
            result = page.evaluate(script)
        except PlaywrightTimeoutError:
            raise ValueError(
                "JavaScript execution timed out. "
                "Simplify the script or check for infinite loops."
            )
        except Exception as e:
            if _is_target_closed(e):
                self._handle_page_crash(page_id)
            error_str = str(e)
            if "SyntaxError" in error_str:
                raise ValueError(
                    f"JavaScript syntax error: {error_str}. "
                    f"Check for invalid syntax. Do not use bare 'return' statements."
                )
            raise
        self._extracted_elements.pop(page_id, None)
        if self.debug_mode:
            result_preview = repr(result)[:500] if result is not None else "None"
            print(f"[DEBUG][BrowserTools.execute_javascript] Result: {result_preview}")
        return result

    def get_page_summary(self, page_id: str, max_elements: int = 50):
        """Return a list of interactive elements visible on the page (up to max_elements, default 50). Each element has: index, selector, text, tag, category, parent_context, type, href. Use the index with click_element(page_id, index) or fill_element(page_id, index, value) to interact. Call this after ANY action that changes the DOM to get fresh data."""
        if self.debug_mode:
            print(f"[DEBUG][BrowserTools.get_page_summary] page_id={page_id}")
        page = self._get_page(page_id)
        if self.debug_mode:
            print(
                f"[DEBUG][BrowserTools.get_page_summary] Extracting elements from page url={page.url}"
            )
        try:
            full_elements = self.extractor.extract_elements(page)
        except Exception as e:
            if _is_target_closed(e):
                self._handle_page_crash(page_id)
            raise
        self._extracted_elements[page_id] = full_elements
        if self.debug_mode:
            print(
                f"[DEBUG][BrowserTools.get_page_summary] Extracted {len(full_elements)} elements"
            )
            for elem in full_elements:
                print(
                    f"[DEBUG][BrowserTools.get_page_summary]   [{elem['index']}] <{elem['tag']}> text={elem['text'][:50]!r} bbox={elem['viewport_bbox']} selector={elem['selector'][:80]}"
                )
        slim = [_slim_element(e) for e in full_elements]
        total = len(slim)
        if total > max_elements:
            slim = slim[:max_elements]
            slim.append(
                {
                    "note": (
                        f"{total - max_elements} more elements not shown (total: {total}). "
                        f"click_element/fill_element still work with any index 0-{total - 1}. "
                        f"Call get_page_summary(page_id, max_elements={total}) to see all."
                    )
                }
            )
        return slim

    def get_element_by_summary_index(self, page_id: str, index: int):
        """Get detailed information about a specific element by its index from the last get_page_summary result. Returns the full element data including selector, text, tag, bbox, attributes. Use this when you need details beyond what get_page_summary shows (e.g., coordinates, attributes)."""
        if self.debug_mode:
            print(
                f"[DEBUG][BrowserTools.get_element_by_summary_index] page_id={page_id}, index={index}"
            )
        element = self._resolve_element(page_id, index)
        if self.debug_mode:
            print(
                f"[DEBUG][BrowserTools.get_element_by_summary_index] Found element: <{element['tag']}> text={element['text'][:50]!r}"
            )
        return element

    def screenshot(
        self,
        page_id: str,
        show_markers: bool = False,
        highlight_element: Optional[str] = None,
    ) -> Path:
        """Take a screenshot of the page for visual analysis. Use show_markers=True to overlay numbered labels on all interactive elements (buttons, links, inputs). Use highlight_element with a CSS selector to highlight a specific element with a red border."""
        if self.debug_mode:
            print(
                f"[DEBUG][BrowserTools.screenshot] page_id={page_id}, show_markers={show_markers}, highlight_element={highlight_element}"
            )
        page = self._get_page(page_id)

        if self.debug_mode:
            print(
                f"[DEBUG][BrowserTools.screenshot] Taking screenshot of url={page.url}"
            )

        if show_markers:
            try:
                full_elements = self.extractor.extract_elements(page)
            except Exception as e:
                if _is_target_closed(e):
                    self._handle_page_crash(page_id)
                raise
            self._extracted_elements[page_id] = full_elements
            if self.debug_mode:
                print(
                    f"[DEBUG][BrowserTools.screenshot] Injecting markers for {len(full_elements)} elements"
                )
            marker_js = """
            (elements) => {
                const container = document.createElement('div');
                container.id = '__abstra_markers__';
                container.style.pointerEvents = 'none';
                container.style.position = 'absolute';
                container.style.top = '0';
                container.style.left = '0';
                container.style.zIndex = '2147483647';
                document.body.appendChild(container);

                elements.forEach((elem) => {
                    const border = document.createElement('div');
                    border.style.position = 'absolute';
                    border.style.left = elem.bbox.x + 'px';
                    border.style.top = elem.bbox.y + 'px';
                    border.style.width = elem.bbox.width + 'px';
                    border.style.height = elem.bbox.height + 'px';
                    border.style.border = '2px solid rgba(255, 0, 0, 0.7)';
                    border.style.borderRadius = '2px';
                    border.style.pointerEvents = 'none';
                    container.appendChild(border);

                    const label = document.createElement('div');
                    label.textContent = elem.index;
                    label.style.position = 'absolute';
                    label.style.left = elem.bbox.x + 'px';
                    label.style.top = (elem.bbox.y - 24) + 'px';
                    label.style.background = 'rgba(255, 0, 0, 0.85)';
                    label.style.color = 'white';
                    label.style.fontSize = '16px';
                    label.style.fontWeight = 'bold';
                    label.style.fontFamily = 'monospace';
                    label.style.padding = '2px 6px';
                    label.style.borderRadius = '4px';
                    label.style.lineHeight = '20px';
                    label.style.pointerEvents = 'none';
                    label.style.whiteSpace = 'nowrap';
                    container.appendChild(label);
                });
            }
            """
            page.evaluate(marker_js, full_elements)

        highlight_css = None
        if highlight_element:
            safe_selector = highlight_element.replace("\\", "\\\\").replace('"', '\\"')
            highlight_css = f"{safe_selector} {{ outline: 3px solid red !important; box-shadow: 0 0 8px rgba(255,0,0,0.6) !important; }}"
            if self.debug_mode:
                print(
                    f"[DEBUG][BrowserTools.screenshot] Highlight CSS for selector: {highlight_element}"
                )

        screenshot_kwargs: Dict[str, Any] = {"type": "jpeg", "full_page": show_markers}
        if highlight_css:
            screenshot_kwargs["style"] = highlight_css

        try:
            image_data = page.screenshot(**screenshot_kwargs)
        except Exception as e:
            if _is_target_closed(e):
                self._handle_page_crash(page_id)
            raise

        if self.debug_mode:
            print(
                f"[DEBUG][BrowserTools.screenshot] Screenshot taken, size={len(image_data)} bytes"
            )

        if show_markers:
            page.evaluate(
                "() => { const el = document.getElementById('__abstra_markers__'); if (el) el.remove(); }"
            )
            if self.debug_mode:
                print("[DEBUG][BrowserTools.screenshot] Marker overlays removed")

        with NamedTemporaryFile(delete=False, suffix=".jpg") as f:
            image_path = Path(f.name)
            if self.debug_mode:
                print(f"[DEBUG][BrowserTools.screenshot] Saving to {image_path}")
            image_path.write_bytes(image_data)

        return image_path

    def __tools__(self):
        tools = [
            self.navigate_to_url.__name__,
            self.list_pages.__name__,
            self.click.__name__,
            self.click_element.__name__,
            self.fill.__name__,
            self.fill_element.__name__,
            self.get_html.__name__,
            self.get_text.__name__,
            self.get_page_summary.__name__,
            self.get_element_by_summary_index.__name__,
            self.get_attribute.__name__,
            self.get_attributes.__name__,
            self.get_all_links.__name__,
            self.execute_javascript.__name__,
            self.wait.__name__,
            self.screenshot.__name__,
        ]

        if self.allow_close_page:
            tools.append(self.close_page.__name__)
        if self.listen_network:
            tools.append(self.get_network_requests.__name__)
        if self.listen_console:
            tools.append(self.get_console_logs.__name__)

        return tools



import os

import requests

# LOGS
DEFAULT_LOGLEVEL = "WARNING"
LOGLEVEL = lambda: os.getenv("ABSTRA_LOGLEVEL", DEFAULT_LOGLEVEL)  # noqa: E731
NOISY_LOGLEVEL = lambda: os.getenv("ABSTRA_NOISY_LOGLEVEL", DEFAULT_LOGLEVEL)  # noqa: E731

PROCESS_LOGFORMAT = "[%(asctime)s][%(levelname)s][%(name)s][%(process)d]%(message)s"
DEFAULT_LOGFORMAT = "[%(asctime)s][%(levelname)s][%(name)s] %(message)s"
LOGFORMAT = lambda: os.getenv("ABSTRA_LOGFORMAT", DEFAULT_LOGFORMAT)  # noqa: E731
CLOUD_SAMPLE_RATE = float(os.getenv("ABSTRA_CLOUD_SAMPLE_RATE", 0.2))
LOCAL_SAMPLE_RATE = float(os.getenv("ABSTRA_LOCAL_SAMPLE_RATE", 0.0))

# SERVER
HOST = os.getenv("ABSTRA_HOST", "localhost")
DEFAULT_PORT = os.getenv("PORT") or os.getenv("ABSTRA_SERVER_PORT")

# PRODUCTION ENVIRONMENT
BUILD_ID = os.getenv("ABSTRA_BUILD_ID") or "dev"
PROJECT_ID = os.getenv("ABSTRA_PROJECT_ID") or "dev-project-id"
PROJECT_URL = os.getenv("ABSTRA_PROJECT_URL")

# PASSWORDLESS AUTHENTICATION
EMAIL_JWT_AUDIENCE = f"abstra:email:{PROJECT_ID}"
PUBLIC_KEY = os.getenv("ABSTRA_JWT_PUBLIC_KEY_PEM")

# OIDC AUTHENTICATION
OIDC_CLIENT_ID = lambda: os.getenv("ABSTRA_OIDC_CLIENT_ID")  # noqa: E731
OIDC_AUTHORITY = lambda: os.getenv("ABSTRA_OIDC_AUTHORITY")  # noqa: E731


# CLOUD API
CLOUD_API_DEFAULT_ENDPOINT = "https://cloud-api.abstra.cloud"
CLOUDFRONT_CLOUD_API_ENDPOINT = "https://cloud.abstra.io/api/cloud-api"


def select_cloud_api_endpoint() -> str:
    try:
        response = requests.get(f"{CLOUD_API_DEFAULT_ENDPOINT}/healthcheck", timeout=10)
        if response.status_code == 200:
            return CLOUD_API_DEFAULT_ENDPOINT
    except Exception as e:
        print(f"Healthcheck failed for {CLOUD_API_DEFAULT_ENDPOINT}: {e}")

    return CLOUDFRONT_CLOUD_API_ENDPOINT


CLOUD_API_ENDPOINT = os.getenv("CLOUD_API_ENDPOINT") or select_cloud_api_endpoint()


CLOUD_API_CLI_URL = f"{CLOUD_API_ENDPOINT}/cli"

CLOUD_API_PROD_SHARED_TOKEN = os.getenv("ABSTRA_CLOUD_API_SHARED_TOKEN", "shared")
CLOUD_API_PROD_HEADERS = {"shared-token": CLOUD_API_PROD_SHARED_TOKEN}
CLOUD_API_PROD_URL = f"{CLOUD_API_ENDPOINT}/apps"


# DOCS

DOCS_URL = os.getenv("ABSTRA_DOCS_URL", "https://www.abstra.io/docs")

# DEBUG
DISABLE_STDIO_PATCH = os.getenv("ABSTRA_DISABLE_STDIO_PATCH", "false") == "true"
REQUEST_TIMEOUT = int(os.getenv("ABSTRA_REQUEST_TIMEOUT", 300))
MAX_HTTP_CLIENT_THREADS = int(os.getenv("ABSTRA_MAX_HTTP_CLIENT_THREADS", 10))

# FILES
FILES_FOLDER = os.getenv("ABSTRA_FILES_FOLDER")
DISABLED_STAGES_FOLDER = os.getenv("ABSTRA_DISABLED_STAGES_FOLDER")

# WEBEDITOR
WAITING_ROOM_URL = os.getenv("ABSTRA_WAITING_ROOM_URL") or ""

# ENVIRONMENT
IS_PRODUCTION = os.getenv("ABSTRA_ENVIRONMENT") == "production"
IS_DEVELOPMENT = not IS_PRODUCTION
EDITOR_MODE = os.getenv("ABSTRA_EDITOR_MODE") or "local"

# FORMS CONFIG
SHOW_WATERMARK = os.getenv("ABSTRA_SHOW_WATERMARK", "false") == "true"

# AMQP
WORKER_LOG_TO_QUEUE = os.getenv("ABSTRA_WORKER_LOG_TO_QUEUE", "false") == "true"
RABBITMQ_EXECUTION_QUEUE = os.getenv("ABSTRA_RABBITMQ_EXECUTION_QUEUE", "executions")
RABBITMQ_DEFAUT_EXCHANGE = os.getenv("ABSTRA_RABBITMQ_DEFAUT_EXCHANGE", "")
RABBITMQ_CONNECTION_URI = os.getenv("ABSTRA_RABBITMQ_CONNECTION_URI")

# AMQP Connection Resilience
RABBITMQ_CONNECTION_TIMEOUT_SECONDS = float(
    os.getenv("ABSTRA_RABBITMQ_CONNECTION_TIMEOUT_SECONDS", 60)
)
RABBITMQ_RETRY_MAX_ATTEMPTS = int(os.getenv("ABSTRA_RABBITMQ_RETRY_MAX_ATTEMPTS", 5))
RABBITMQ_RETRY_INITIAL_DELAY_SECONDS = float(
    os.getenv("ABSTRA_RABBITMQ_RETRY_INITIAL_DELAY_SECONDS", 2)
)

# WORKER PROCESSING
PROCESS_TIMEOUT_SECONDS = int(os.getenv("ABSTRA_PROCESS_TIMEOUT_SECONDS", 60 * 60 * 2))

ABSTRA_EXECUTOR_POOL_SIZE = int(os.getenv("ABSTRA_EXECUTOR_POOL_SIZE", 2))
ABSTRA_MIN_FORMS_EXECUTORS = int(os.getenv("ABSTRA_MIN_FORMS_EXECUTORS", 0))
ABSTRA_EXECUTOR_MAX_EXECUTIONS = int(os.getenv("ABSTRA_EXECUTOR_MAX_EXECUTIONS", 8))
ABSTRA_EXECUTOR_WARMUP_TIMEOUT = float(
    os.getenv("ABSTRA_EXECUTOR_WARMUP_TIMEOUT", 60.0)
)
ABSTRA_EXECUTOR_WARMUP_PARALLELISM = int(
    os.getenv("ABSTRA_EXECUTOR_WARMUP_PARALLELISM", 1)
)
ABSTRA_EXECUTOR_ACQUIRE_TIMEOUT = float(
    os.getenv("ABSTRA_EXECUTOR_ACQUIRE_TIMEOUT", 600.0)  # 10 minutes default
)

# DEBUG MODE
ABSTRA_DEBUG_MODE_INTERVAL = int(os.getenv("ABSTRA_DEBUG_MODE_INTERVAL", 30))

# GUNICORN SERVER
WORKERS = os.getenv("ABSTRA_WORKERS", 2)
THREADS = os.getenv("ABSTRA_THREADS", 20)
WORKER_CLASS = os.getenv("ABSTRA_WORKER_CLASS", "gthread")
WORKER_CONNECTIONS = int(os.getenv("ABSTRA_WORKER_CONNECTIONS", 1000))
WORKER_TEMP_DIR = os.getenv("ABSTRA_WORKER_TEMP_DIR")

# GIT
FORCE_GIT_CLIENT = os.getenv("ABSTRA_FORCE_GIT_CLIENT")
REMOTE_GIT_URL = f"{CLOUD_API_CLI_URL}/git/repo.git"
REMOTE_NAME = "abstra"



import ast
import io
import pathlib
import shutil
import sys
import tempfile
import time
import typing as t
import uuid
import warnings
from pathlib import Path
from typing import Generator, Optional, Set, Union

from werkzeug.datastructures import FileStorage

from abstra_internals.utils.ast_cache import ASTCache
from abstra_internals.utils.fs_cache import get_cached_cwd, get_path_cache
from abstra_internals.utils.platform import is_windows

FILE_TYPES = {
    "txt": {
        "mime": "text/plain",
        "extensions": ["txt"],
    },
    "md": {
        "mime": "text/markdown",
        "extensions": ["md"],
    },
    "csv": {
        "mime": "text/csv",
        "extensions": ["csv"],
    },
    "html": {
        "mime": "text/html",
        "extensions": ["html"],
    },
    "css": {
        "mime": "text/css",
        "extensions": ["css"],
    },
    "py": {
        "mime": "text/x-python-script",
        "extensions": ["py"],
    },
    "png": {
        "mime": "image/png",
        "extensions": ["png"],
    },
    "jpg": {
        "mime": "image/jpeg",
        "extensions": ["jpg", "jpeg"],
    },
    "jpeg": {
        "mime": "image/jpeg",
        "extensions": ["jpg", "jpeg"],
    },
    "mp3": {
        "mime": "audio/mp3",
        "extensions": ["mp3"],
    },
    "wav": {
        "mime": "audio/wav",
        "extensions": ["wav"],
    },
    "flac": {
        "mime": "audio/flac",
        "extensions": ["flac"],
    },
    "aac": {
        "mime": "audio/aac",
        "extensions": ["aac"],
    },
    "ogg": {
        "mime": "audio/ogg",
        "extensions": ["ogg"],
    },
    "wma": {
        "mime": "audio/x-ms-wma",
        "extensions": ["wma"],
    },
    "avi": {
        "mime": "video/avi",
        "extensions": ["avi"],
    },
    "mp4": {
        "mime": "video/mp4",
        "extensions": ["mp4"],
    },
    "mkv": {
        "mime": "video/x-matroska",
        "extensions": ["mkv"],
    },
    "mov": {
        "mime": "video/quicktime",
        "extensions": ["mov"],
    },
    "webm": {
        "mime": "video/webm",
        "extensions": ["webm"],
    },
    "flv": {
        "mime": "video/x-flv",
        "extensions": ["flv"],
    },
    "wmv": {
        "mime": "video/x-ms-wmv",
        "extensions": ["wmv"],
    },
    "m4v": {
        "mime": "video/x-m4v",
        "extensions": ["m4v"],
    },
    "zip": {
        "mime": "application/zip",
        "extensions": ["zip"],
    },
    "rar": {
        "mime": "application/vnd.rar",
        "extensions": ["rar"],
    },
    "7z": {
        "mime": "application/x-7z-compressed",
        "extensions": ["7z"],
    },
    "tar": {
        "mime": "application/x-tar",
        "extensions": ["tar"],
    },
    "gzip": {
        "mime": "application/gzip",
        "extensions": ["gzip"],
    },
    "js": {
        "mime": "application/javascript",
        "extensions": ["js"],
    },
    "ts": {
        "mime": "application/typescript",
        "extensions": ["ts"],
    },
    "json": {
        "mime": "application/json",
        "extensions": ["json"],
    },
    "xml": {
        "mime": "application/xml",
        "extensions": ["xml"],
    },
    "yaml": {
        "mime": "application/x-yaml",
        "extensions": ["yaml"],
    },
    "toml": {
        "mime": "application/toml",
        "extensions": ["toml"],
    },
    "pdf": {
        "mime": "application/pdf",
        "extensions": ["pdf"],
    },
    "xls": {
        "mime": "application/vnd.ms-excel",
        "extensions": ["xls"],
    },
    "doc": {
        "mime": "application/msword",
        "extensions": ["doc"],
    },
    "ppt": {
        "mime": "application/vnd.ms-powerpoint",
        "extensions": ["ppt"],
    },
    "docx": {
        "mime": "application/vnd.openxmlformats-officedocument.wordprocessingml.document",
        "extensions": ["docx"],
    },
    "pptx": {
        "mime": "application/vnd.openxmlformats-officedocument.presentationml.presentation",
        "extensions": ["pptx"],
    },
    "image": {
        "mime": "image/*",
        "extensions": ["png", "jpg", "jpeg"],
    },
    "video": {
        "mime": "video/*",
        "extensions": ["avi", "mp4", "mkv", "mov", "webm", "flv", "wmv", "m4v"],
    },
    "audio": {
        "mime": "audio/*",
        "extensions": ["mp3", "wav", "flac", "aac", "ogg", "wma"],
    },
    "text": {
        "mime": "text/*",
        "extensions": ["txt", "md", "csv", "html", "css", "py"],
    },
    "application": {
        "mime": "application/*",
        "extensions": [
            "zip",
            "rar",
            "7z",
            "tar",
            "gzip",
            "js",
            "ts",
            "json",
            "xml",
            "yaml",
            "toml",
            "pdf",
            "xls",
            "doc",
            "ppt",
            "docx",
            "pptx",
        ],
    },
    "unknown": {
        "mime": "*",
        "extensions": [
            "txt",
            "md",
            "csv",
            "html",
            "css",
            "py",
            "png",
            "jpg",
            "jpeg",
            "mp3",
            "wav",
            "flac",
            "aac",
            "ogg",
            "wma",
            "avi",
            "mp4",
            "mkv",
            "mov",
            "webm",
            "flv",
            "wmv",
            "m4v",
            "zip",
            "rar",
            "7z",
            "tar",
            "gzip",
            "js",
            "ts",
            "json",
            "xml",
            "yaml",
            "toml",
            "pdf",
            "xls",
            "doc",
            "ppt",
            "docx",
            "pptx",
            "image",
            "video",
            "audio",
            "text",
            "application",
            "unknown",
            "zip",
            "rar",
            "7z",
            "tar",
            "gzip",
            "js",
            "ts",
            "json",
            "xml",
            "yaml",
            "toml",
            "pdf",
            "xls",
            "doc",
            "ppt",
            "docx",
            "pptx",
        ],
    },
}


def is_relative_path(path: str) -> bool:
    if path.startswith("<"):
        return False

    from pathlib import Path

    try:
        return not Path(path).is_absolute()
    except (OSError, ValueError):
        return not path.startswith("/") and not path.startswith("<")


def path2module(path: Path) -> str:
    if path.suffix == ".py":
        path = path.with_suffix("")
    if path.name == "__init__":
        path = path.parent
    return str(path).replace("/", ".").replace("\\", ".")


def module2path(module: str, package: bool) -> Path:
    path_list = module.split(".")

    # Handle trailing dots
    is_first_trailing_dot = True
    for index, path in enumerate(path_list):
        if path.__len__() == 0:
            if is_first_trailing_dot:
                path_list[index] = "."
                is_first_trailing_dot = False
            else:
                path_list[index] = ".."
        else:
            break

    if package:
        path_list.append("__init__.py")
    else:
        path_list[-1] += ".py"
    return Path(*path_list)


def _get_file_path_from_absolute_module(module: str) -> Union[Path, None]:
    module_path = Path(module2path(module, package=False))
    if module_path.is_file():
        return module_path

    package_path = Path(module2path(module, package=True))
    if package_path.is_file():
        return package_path

    return None


def _get_file_path_from_relative_module(
    module: str, parent_path: Path
) -> Union[Path, None]:
    module_path = parent_path / module2path(module, package=False)
    resolved_module_path = (
        get_path_cache().get_resolved_path(module_path).relative_to(get_cached_cwd())
    )

    if resolved_module_path.is_file():
        return resolved_module_path

    package_path = parent_path / module2path(module, package=True)
    resolved_package_path = (
        get_path_cache().get_resolved_path(package_path).relative_to(get_cached_cwd())
    )

    if resolved_package_path.is_file():
        return resolved_package_path

    return None


def _get_file_path_depending_on_node_level(
    node_level: int, module: str, path_parent: Path
) -> Optional[Path]:
    return (
        _get_file_path_from_relative_module(module, path_parent)
        if node_level > 0
        else _get_file_path_from_absolute_module(module)
    )


def _get_file_path(
    node_level: int,
    module: Optional[str],
    alias_name: str,
    path_parent: Path,
) -> Optional[Path]:
    prefix = "." * (node_level)
    has_module = module is not None and module != ""

    # Check if it's importing a file
    module_name = f"{module}." if has_module else ""
    relative_module = f"{prefix}{module_name}{alias_name}"
    file_path = _get_file_path_depending_on_node_level(
        node_level, relative_module, path_parent
    )

    # Check if it's importing an entity from a file
    if file_path is None:
        relative_module = f"{prefix}{module if has_module else ''}"
        file_path = _get_file_path_depending_on_node_level(
            node_level, relative_module, path_parent
        )

    return file_path


def silent_traverse_code(
    path: Path, raise_on_syntax_errors=False
) -> Generator[Path, None, None]:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        yield from traverse_code(path, raise_on_syntax_errors, set())


_SDK_PACKAGES = frozenset({"abstra", "abstra_internals", "abstra_statics"})


def _is_sdk_module(module_name: str) -> bool:
    """Check if a module belongs to the Abstra SDK.

    SDK modules must not be cleared from sys.modules because they hold
    runtime state (e.g. SDKContextStore) that must persist across user
    script executions within the same process.
    """
    top_level = module_name.split(".")[0]
    return top_level in _SDK_PACKAGES


def clear_local_modules(filepath: Path, root: Path) -> Set[str]:
    """
    Clear all local project modules from sys.modules to ensure fresh imports.

    This is necessary because executor processes persist between executions,
    and Python caches imported modules in sys.modules. When a dependency file
    is modified, we need to ensure the next execution loads the fresh version
    from disk rather than using the stale cached version.

    SDK modules (abstra, abstra_internals, abstra_statics) are never cleared
    because they hold runtime state that must survive between executions.

    Args:
        filepath: The entrypoint file to traverse for dependencies.
        root: The project root path.

    Returns:
        The set of module names that were cleared.
    """
    resolved_root = root.resolve()
    modules_to_remove: Set[str] = set()

    try:
        dependencies = list(silent_traverse_code(filepath))
    except Exception:
        return modules_to_remove

    for dep_path in dependencies:
        try:
            # Construct module name from path relative to project root
            # Always resolve to handle symlinks consistently (e.g., /var -> /private/var on macOS)
            if dep_path.is_absolute():
                resolved = dep_path.resolve()
            else:
                resolved = (resolved_root / dep_path).resolve()

            relative_path = resolved.relative_to(resolved_root)
            if relative_path.suffix == ".py":
                dotted_name = ".".join(relative_path.with_suffix("").parts)
                modules_to_remove.add(dotted_name)

                parts = relative_path.with_suffix("").parts
                for i in range(1, len(parts)):
                    parent_module = ".".join(parts[:i])
                    modules_to_remove.add(parent_module)

        except (ValueError, OSError):
            continue

    cleared = set()
    for module_name in modules_to_remove:
        if _is_sdk_module(module_name):
            continue
        if module_name in sys.modules:
            del sys.modules[module_name]
            cleared.add(module_name)

    return cleared


def traverse_code(
    path: Path, raise_on_syntax_errors=False, yielded_files: Optional[Set[Path]] = None
) -> Generator[Path, None, None]:
    if yielded_files is None:
        yielded_files = set()

    try:
        parsed = ASTCache.get(path)

        for node in parsed.body:
            if isinstance(node, ast.Import):
                for alias in node.names:
                    file_path = _get_file_path_from_absolute_module(alias.name)

                    if (
                        file_path is not None
                        and file_path not in yielded_files
                        and file_path.absolute() != path.absolute()
                    ):
                        yield from traverse_code(
                            file_path, raise_on_syntax_errors, yielded_files
                        )

            if isinstance(node, ast.ImportFrom):
                for alias in node.names:
                    file_path = _get_file_path(
                        node.level, node.module, alias.name, path.parent
                    )

                    if (
                        file_path is not None
                        and file_path not in yielded_files
                        and file_path.absolute() != path.absolute()
                    ):
                        yield from traverse_code(
                            file_path, raise_on_syntax_errors, yielded_files
                        )

    except Exception as e:
        if raise_on_syntax_errors and isinstance(e, SyntaxError):
            raise e

    if path not in yielded_files:
        yield path
        yielded_files.add(path)


def get_tmp_upload_dir():
    uploaded_files_dir = Path(tempfile.gettempdir(), "_uploaded_files")
    uploaded_files_dir.mkdir(exist_ok=True)
    return uploaded_files_dir


def make_random_tmp_path(name: Optional[str] = None) -> Path:
    if name is None:
        name = str(uuid.uuid4())
    else:
        name = str(uuid.uuid4()) + "/" + Path(name).name
    path = get_tmp_upload_dir() / name
    path.parent.mkdir(exist_ok=True)
    return path


def upload_file(
    input: t.Union[str, FileStorage, io.IOBase, pathlib.Path],
    custom_name: Optional[str] = None,
) -> str:
    def get_public_path(path: Path) -> str:
        # Get path relative to temp upload dir, which includes uuid/filename
        relative_path = path.relative_to(get_tmp_upload_dir())
        return f"/_files/{relative_path}"

    if isinstance(input, str):
        if input.startswith("http") or input.startswith("data:"):
            return input

        path = make_random_tmp_path(custom_name or input)
        shutil.copy(input, path)
        return get_public_path(path)

    if isinstance(input, (io.BufferedReader, io.TextIOWrapper)):
        path = make_random_tmp_path(custom_name or input.name)
        shutil.copy(input.name, path)
        return get_public_path(path)

    if isinstance(input, FileStorage):
        path = make_random_tmp_path(custom_name or input.filename)
        input.save(path)
        return get_public_path(path)

    if isinstance(input, pathlib.Path):
        path = make_random_tmp_path(custom_name or input.name)
        shutil.copy(input, path)
        return get_public_path(path)

    if isinstance(input, io.IOBase):
        path = make_random_tmp_path(custom_name)
        with open(path, "wb") as f:
            shutil.copyfileobj(input, f)
        return get_public_path(path)

    raise ValueError(f"Cannot upload {type(input)}")


CONFLICTING_STATIC_PATHS = [
    # Player routes
    "oidc-login-callback",
    "oidc-logout-callback",
    "login",
    "threads",
]

CONFLICTING_DYNAMIC_PATHS = [
    # Player routes
    "error/",
    "_player/",
    "_page/",
    "form/",
    "forms/",
    "hook/",
    "hooks/",
    "page/",
    "pages/",
    # Editor routes
    "_editor/",
]


def is_path_in_conflict(path: str):
    return path in CONFLICTING_STATIC_PATHS or any(
        path.startswith(p) for p in CONFLICTING_DYNAMIC_PATHS
    )


def generate_conflictless_path(path: str):
    if is_path_in_conflict(path):
        random_suffix = str(uuid.uuid4())[:8]
        return f"{path}-{random_suffix}"
    else:
        return path


def extension_to_mime(extension: str) -> str:
    extension_without_dot = extension.replace(".", "")
    if extension_without_dot in FILE_TYPES:
        return FILE_TYPES[extension_without_dot]["mime"]
    return FILE_TYPES["unknown"]["mime"]


def is_file_locked(filepath: Path) -> bool:
    if not filepath.exists():
        return False

    try:
        if is_windows():
            with open(filepath, "a", encoding="utf-8"):
                pass

            with open(filepath, "r", encoding="utf-8"):
                pass
        else:
            with open(filepath, "r+", encoding="utf-8"):
                pass
        return False
    except (PermissionError, OSError):
        return True


def wait_for_file_unlock(
    filepath: Path, timeout: float = 2.0, check_interval: float = 0.05
) -> bool:
    if not filepath.exists():
        return True

    start_time = time.time()
    while time.time() - start_time < timeout:
        if not is_file_locked(filepath):
            return True
        time.sleep(check_interval)

    return False


def safe_read_file(filepath: Path, timeout: float = 2.0) -> Optional[str]:
    if not filepath.is_file():
        return None

    if not wait_for_file_unlock(filepath, timeout):
        return None

    try:
        return filepath.read_text(encoding="utf-8")
    except (PermissionError, OSError, UnicodeDecodeError) as e:
        print(f"Error reading {filepath}: {e}")
        return None


def safe_write_file(filepath: Path, content: str, timeout: float = 2.0) -> bool:
    if not wait_for_file_unlock(filepath, timeout):
        return False

    try:
        filepath.write_text(content, encoding="utf-8")
        return True
    except (PermissionError, OSError) as e:
        print(f"Error writing {filepath}: {e}")
        return False



from pathlib import Path
from typing import List

from abstra_internals.controllers.language_server import get_diagnostics
from abstra_internals.repositories.linter.models import LinterIssue, LinterRule
from abstra_internals.repositories.project.project import LocalProjectRepository


class TypeCheckIssue(LinterIssue):
    def __init__(self, filename: str, messages: List[str]) -> None:
        super().__init__()
        count = len(messages)
        lines = [f"{filename}: {count} issue{'s' if count != 1 else ''}"]
        lines.extend(messages)
        self.label = "\n".join(lines)
        self.fixes = []


class TypeCheckingRule(LinterRule):
    label = "Type checking"
    type = "info"
    fix_with_ai = True

    def find_issues(self) -> List[LinterIssue]:
        project = LocalProjectRepository().load()
        issues: List[LinterIssue] = []

        for entrypoint in project.iter_entrypoints():
            path = Path(entrypoint)
            if not path.is_file() or path.suffix != ".py":
                continue

            try:
                code = path.read_text(encoding="utf-8")
            except (OSError, UnicodeDecodeError):
                continue

            diagnostics = get_diagnostics(code)
            messages = []
            for diag in diagnostics:
                if diag.get("severity", 1) > 2:
                    continue
                line = diag.get("range", {}).get("start", {}).get("line", 0) + 1
                message = diag.get("message", "Type error")
                messages.append(f"  Line {line}: {message}")
            if messages:
                issues.append(TypeCheckIssue(str(path), messages))

        return issues



import threading
from abc import ABC, abstractmethod
from typing import List

from abstra_internals.repositories.linter.models import LinterCheck
from abstra_internals.repositories.linter.rules import rules


class LinterRepository(ABC):
    @abstractmethod
    def find_issues_in_codebase(self) -> List[LinterCheck]:
        pass

    @abstractmethod
    def update_checks(self) -> List[LinterCheck]:
        pass

    @abstractmethod
    def fix_issue_in_codebase(self, rule_name: str, fix_name: str) -> bool:
        pass

    @abstractmethod
    def fix_all_linters(self):
        pass

    @abstractmethod
    def get_blocking_checks(self) -> List[LinterCheck]:
        pass


def check_rule(rule, checks_list):
    check = rule.check()
    checks_list.append(check)


LINTER_TYPE_PRIORITY = {"security": 0, "bug": 1, "warning": 2, "info": 3}


class LocalLinterRepository(LinterRepository):
    def __init__(self):
        self.checks: List[LinterCheck] = []

    def find_issues_in_codebase(self) -> List[LinterCheck]:
        """
        Retrieve all linter checks that have been performed on the project.

        This method returns the cached list of all linter checks from the most recent
        analysis run. Each check represents a specific rule that was evaluated against
        the project code, including any issues found and available fixes.

        Returns:
            List[LinterCheck]: List of linter check objects, each containing:
                - name: Unique identifier for the linter rule
                - type: Severity level ('info', 'warning', 'error', 'security', 'bug')
                - description: Human-readable description of what the rule checks
                - issues: List of specific issues found by this rule
                - fixes: Available automatic fixes for the issues
            ```

        Note:
            - Returns cached results from the last update_checks() call
            - Empty list if update_checks() has never been called
            - Issues include file location and line number information
            - Fixes can be applied automatically using fix_linter() method
            - Check types determine severity and whether they block deployment

        Copywritings:
            Retrieve all linter checks
            Retrieving all linter checks...
        """
        if len(self.checks) == 0:
            self.update_checks()
        return self.checks

    def update_checks(self):
        new_checks = []
        threads = []

        for rule in rules:
            thread = threading.Thread(
                target=check_rule,
                args=(rule, new_checks),
                name=f"LinterCheck[{rule.name}]",
            )
            thread.start()
            threads.append(thread)

        for thread in threads:
            thread.join()

        new_checks.sort(key=lambda c: LINTER_TYPE_PRIORITY.get(c.type, 4))
        self.checks = new_checks

        return self.checks

    def fix_issue_in_codebase(self, rule_name: str, fix_name: str):
        """
        Apply a specific automatic fix for a linter issue.

        This method searches for a specific linter rule and fix combination,
        then applies the fix automatically. It's used to resolve individual
        linter issues without affecting other parts of the code.

        Args:
            rule_name (str): Name of the linter rule that found the issue.
                Must match exactly with the rule name from get_checks().
            fix_name (str): Name of the specific fix to apply.
                Must match exactly with a fix name available for the rule's issues.

        Returns:
            bool: True if the fix was found and applied successfully, False otherwise.

        Example:
            ```python
            linter_repo = LocalLinterRepository()
            linter_repo.update_checks()

            checks = linter_repo.get_checks()

            # Find issues that can be fixed
            for check in checks:
                if check.issues:
                    print(f"Rule: {check.name}")
                    for issue in check.issues:
                        for fix in issue.fixes:
                            print(f"Available fix: {fix.name}")

                            # Apply a specific fix
                            success = linter_repo.fix_linter(check.name, fix.name)
                            if success:
                                print(f"✓ Applied fix: {fix.name}")
                            else:
                                print(f"✗ Failed to apply fix: {fix.name}")

            # Example: Fix a specific formatting issue
            success = linter_repo.fix_linter("code_formatting", "add_missing_imports")
            if success:
                print("Import formatting fixed!")

            # Example: Fix security vulnerability
            success = linter_repo.fix_linter("security_check", "sanitize_input")
            if success:
                print("Security issue resolved!")
            else:
                print("Fix not found or failed to apply")

            # Re-run checks to see if issues were resolved
            linter_repo.update_checks()
            updated_checks = linter_repo.get_checks()
            ```

        Note:
            - Rule and fix names are case-sensitive and must match exactly
            - Returns False if rule_name or fix_name is not found
            - Applies the fix immediately to the project files
            - Some fixes may modify multiple files or lines of code
            - Re-run update_checks() after fixing to see updated results
            - Use get_checks() first to see available rules and fixes

        Copywritings:
            Apply a specific automatic fix for a linter issue
            Applying a specific automatic fix for a linter issue...
        """
        for check in self.checks:
            if check.name == rule_name:
                for issue in check.issues:
                    for fix in issue.fixes:
                        if fix.name == fix_name:
                            fix.fix()
                            self._update_check_for_rule(rule_name)
                            return True
        return False

    def _update_check_for_rule(self, rule_name: str):
        """Re-run the check for a specific rule and update the cache."""
        for rule in rules:
            if rule.name == rule_name:
                new_check = rule.check()
                self.checks = [
                    new_check if check.name == rule_name else check
                    for check in self.checks
                ]
                break

    def fix_all_linters(self):
        for check in self.checks:
            if check.type != "info":
                for issue in check.issues:
                    for fix in issue.fixes:
                        fix.fix()

    def get_blocking_checks(self) -> List[LinterCheck]:
        return [
            check
            for check in self.checks
            if check.type in ["error", "security", "bug"] and check.issues
        ]


class ProductionLinterRepository(LinterRepository):
    """
    This is a dummy repository. Linters are not available in production.
    """

    def find_issues_in_codebase(self) -> List[LinterCheck]:
        raise NotImplementedError("Linters are not available in production.")

    def update_checks(self) -> List[LinterCheck]:
        raise NotImplementedError("Linters are not available in production.")

    def fix_issue_in_codebase(self, rule_name: str, fix_name: str) -> bool:
        raise NotImplementedError("Linters are not available in production.")

    def fix_all_linters(self):
        raise NotImplementedError("Linters are not available in production.")

    def get_blocking_checks(self) -> List[LinterCheck]:
        raise NotImplementedError("Linters are not available in production.")



from .base_migration import Migration


class Migration007(Migration):
    @staticmethod
    def target_version() -> str:
        return "7.0"

    @staticmethod
    def source_version() -> str:
        return "6.0"

    def _migrate(self) -> None:
        iterators = self.data.get("iterators", [])

        for iterator in iterators:
            iterator["item_name"] = "item"



from io import BufferedReader, TextIOWrapper
from pathlib import Path
from typing import List, Optional, Union

from abstra_internals.entities.forms.widgets.response_types import FileResponse
from abstra_internals.entities.forms.widgets.widget_base import (
    InputWidget,
    MultipleHandler,
)
from abstra_internals.utils.file import extension_to_mime, upload_file


class FileInput(InputWidget):
    """File upload input widget for handling file uploads.

    Attributes:
        value (Union[FileResponse, List[FileResponse], None]): The uploaded file(s). FileResponse is a dictionary-like object: `{'name': str, 'path': pathlib.Path}`. If `multiple` is True, will be a list, otherwise a single FileResponse object.
    """

    type = "file-input"
    multiple: bool = False
    multiple_handler: MultipleHandler
    value: Union[FileResponse, List[FileResponse], None]

    def __init__(
        self,
        label: str,
        *,
        key: Optional[str] = None,
        required: bool = True,
        hint: Optional[str] = None,
        full_width: bool = False,
        disabled: bool = False,
        max_file_size: Optional[int] = None,
        multiple: bool = False,
        min: Optional[int] = None,
        max: Optional[int] = None,
        accepted_formats: Optional[List[str]] = None,
        errors: Optional[Union[List[str], str]] = None,
    ):
        """Initialize a FileInput widget.

        Args:
            label (str): Text label displayed above the input.
            key (Optional[str]): Identifier for the widget, defaults to label if not provided.
            required (bool): Whether a file must be uploaded before proceeding.
            hint (Optional[str]): Help text displayed below the input.
            full_width (bool): Whether the input should take up the full width of its container.
            disabled (bool): Whether the input is non-interactive.
            max_file_size (Optional[int]): Maximum file size in bytes.
            multiple (bool): Whether multiple files can be uploaded.
            min (Optional[int]): Minimum number of files required when multiple=True.
            max (Optional[int]): Maximum number of files allowed when multiple=True.
            accepted_formats (Optional[List[str]]): List of accepted file extensions.
            errors (Optional[Union[List[str], str]]): Pre-defined validation error messages to display.
        """
        self.label = label
        self.key = key or label
        self.required = required
        self.hint = hint
        self.full_width = full_width
        self.disabled = disabled
        self.max_file_size = max_file_size
        self.multiple = multiple
        self.min = min
        self.max = max
        self.accepted_formats = accepted_formats
        self.multiple_handler = MultipleHandler(
            self.multiple, self.min, self.max, self.required
        )
        self.value = None
        self.errors = errors

    def _render(self):
        return {
            "type": self.type,
            "key": self.key,
            "hint": self.hint,
            "label": self.label,
            "value": self._serialize_value(),
            "required": self.required,
            "multiple": self.multiple,
            "min": self.min,
            "max": self.max,
            "acceptedFormats": self.accepted_formats,
            "acceptedMimeTypes": self.get_mime_types,
            "fullWidth": self.full_width,
            "disabled": self.disabled,
            "maxFileSize": self.max_file_size,
            "errors": self.errors,
        }

    @staticmethod
    def __get_file_uri(
        value: Union[FileResponse, str, BufferedReader, TextIOWrapper],
    ) -> str:
        if isinstance(value, str):
            if Path(value).is_file():
                return upload_file(open(value, encoding="utf-8"))
            else:
                return value
        if isinstance(value, FileResponse):
            return value._url
        if isinstance(value, BufferedReader):
            return upload_file(value)
        if isinstance(value, TextIOWrapper):
            return upload_file(value)
        return ""

    def _serialize_value(self) -> List[str]:
        values_list = self.multiple_handler.value_to_list(self.value)
        return [FileInput.__get_file_uri(item) for item in values_list]

    def _parse_value(
        self, value: Optional[List[str]]
    ) -> Union[FileResponse, List[FileResponse], None]:
        file_responses = [FileResponse(url) for url in value or []]
        return self.multiple_handler.value_to_list_or_value(file_responses)

    def _run_validators(self) -> List[str]:
        errors = []
        errors.extend(self.multiple_handler.validate(self.value))
        if not self.value:
            return errors
        if isinstance(self.value, list):
            for file in self.value:
                errors.extend(self._validate_file_extension(file))
        else:
            errors.extend(self._validate_file_extension(self.value))
        return errors

    def _remove_dots(self, accepted_formats: List[str]) -> List[str]:
        return [format.replace(".", "") for format in accepted_formats]

    def _validate_file_extension(self, file: FileResponse) -> List[str]:
        if not self.accepted_formats:
            return []
        file_extension = file._url.split(".")[-1]
        if file_extension not in self._remove_dots(self.accepted_formats):
            return ["i18n_error_invalid_file_format"]
        return []

    @property
    def get_mime_types(self) -> str:
        if not self.accepted_formats:
            return "*"
        return ",".join(extension_to_mime(ext) for ext in self.accepted_formats)



from typing import Any, Protocol


class ConnectionProtocol(Protocol):
    """Protocol for connection-like objects that support bidirectional communication."""

    def send(self, obj: Any) -> None:
        """Send an object through the connection."""
        ...

    def recv(self) -> Any:
        """Receive an object from the connection."""
        ...

    def poll(self, timeout: float = 0.0) -> bool:
        """Check if there are messages available to receive."""
        ...

    def close(self) -> None:
        """Close the connection."""
        ...

    @property
    def closed(self) -> bool:
        """Check if the connection is closed."""
        ...



import json
import mimetypes
from pathlib import Path

import flask
import flask_sock
import jwt as pyjwt

from abstra_internals.constants import get_public_dir
from abstra_internals.controllers.execution.drain import (
    drain_until_response,
    normalize_response,
)
from abstra_internals.controllers.main import MainController
from abstra_internals.entities.execution_context import (
    FormContext,
    HookContext,
    JobContext,
    PageContext,
    Response,
    extract_flask_request,
)
from abstra_internals.environment import (
    BUILD_ID,
    CLOUD_API_PROD_SHARED_TOKEN,
    EDITOR_MODE,
    IS_PRODUCTION,
    OIDC_AUTHORITY,
    OIDC_CLIENT_ID,
    SHOW_WATERMARK,
)
from abstra_internals.logger import AbstraLogger
from abstra_internals.server.cache.control import Cache
from abstra_internals.server.guards.role_guard import (
    Guard,
    PathArgSelector,
    QueryArgSelector,
)
from abstra_internals.server.routes import access_control as ac_router
from abstra_internals.server.routes import auth as auth_router
from abstra_internals.server.routes import workflows as workflows_router
from abstra_internals.server.utils import send_from_dist
from abstra_internals.services.jwt import USER_AUTH_HEADER_KEY
from abstra_internals.settings import Settings
from abstra_internals.usage import player_usage
from abstra_internals.utils import check_is_url, serialize
from abstra_internals.utils.file import get_tmp_upload_dir, path2module, upload_file
from abstra_internals.utils.websockets import bind_ws_with_connection


def get_player_bp(controller: MainController):
    guard = Guard(
        controller.users_repository,
        project_repository=controller.repositories.project,
        enabled=IS_PRODUCTION,
    )
    cache = Cache(enabled=IS_PRODUCTION)

    bp = flask.Blueprint("player", __name__)
    sock = flask_sock.Sock(bp)

    auth_bp = auth_router.get_player_bp(controller)
    bp.register_blueprint(auth_bp, url_prefix="/_auth")

    workflow_bp = workflows_router.get_player_bp(controller)
    bp.register_blueprint(workflow_bp, url_prefix="/_workflows")

    access_control_bp = ac_router.get_player_bp(controller)
    bp.register_blueprint(access_control_bp, url_prefix="/_access-control")

    @bp.route("/_healthcheck")
    def _healthcheck():
        return "ok"

    @bp.get("/_workspace")
    def _get_workspace():
        auth = flask.request.headers.get(USER_AUTH_HEADER_KEY)
        return guard.filtered_workspace(auth).as_dict

    def _resolve_page(path):
        form = controller.get_form_by_path(path)
        if form:
            auth = flask.request.headers.get(USER_AUTH_HEADER_KEY)
            return {
                form.type_name: {
                    **form.browser_runner_dto,
                    "workspace": guard.filtered_workspace(auth).as_dict,
                }
            }

        page = controller.get_page_stage_by_path(path)
        if page:
            auth = flask.request.headers.get(USER_AUTH_HEADER_KEY)
            return {
                "page": {
                    "id": page.id,
                    "path": page.path,
                    "title": page.title,
                    "workspace": guard.filtered_workspace(auth).as_dict,
                }
            }

        flask.abort(404)

    @bp.get("/_pages-home")
    @player_usage
    def _get_home_page():
        return _resolve_page("")

    @bp.get("/_pages/<string:path>")
    @guard.by(PathArgSelector("path"))
    @player_usage
    def _get_page(path):
        return _resolve_page(path)

    @bp.get("/_version")
    def _get_version():
        return BUILD_ID

    @bp.get("/_settings")
    def _get_settings():
        return flask.jsonify(
            {
                "show_watermark": SHOW_WATERMARK,
                "oidc_authority": OIDC_AUTHORITY(),
                "oidc_client_id": OIDC_CLIENT_ID(),
                "editor_mode": EDITOR_MODE,
                "is_production": IS_PRODUCTION,
            }
        )

    @bp.get("/_infra/worker-capacity")
    def _get_worker_capacity():
        capacity = controller.repositories.infra.get_worker_capacity()
        if capacity is None:
            return flask.jsonify(
                {"currentWorkers": 0, "maxWorkers": 0, "isAtCapacity": False}
            )
        return flask.jsonify(
            {
                "currentWorkers": capacity.current_workers,
                "maxWorkers": capacity.max_workers,
                "isAtCapacity": capacity.is_at_capacity,
            }
        )

    @sock.route("/_socket")
    @guard.socket_by(QueryArgSelector("id"))
    def _websocket(ws: flask_sock.Server):
        context = FormContext(request=extract_flask_request(flask.request))
        connection = None

        try:
            id = flask.request.args.get("id")
            if id is None:
                return

            form = controller.get_form(id)
            if not form:
                return

            connection = controller.repositories.producer.enqueue(id, context)
            bind_ws_with_connection(ws, connection, block=True)
        except Exception as e:
            AbstraLogger.capture_exception(e)
        finally:
            if connection is not None:
                try:
                    connection.close()
                except Exception as e:
                    AbstraLogger.capture_exception(e)
            ws.close(message="Done")

    @bp.put("/_files")
    @player_usage
    def _upload_file():
        files = flask.request.files
        if len(files) == 0:
            flask.abort(400)

        return [upload_file(file) for file in files.values()]

    @bp.get("/_files/<path:path>")
    def _get_file(path):
        return flask.send_from_directory(get_tmp_upload_dir(), path)

    @bp.get("/_public/<path:path>")
    def _get_public_file(path):
        return flask.send_from_directory(get_public_dir(), path)

    @bp.get("/_assets/favicon.ico")
    @cache.assets()
    def _favicon():
        favicon_path = controller.get_workspace().favicon_url
        if not favicon_path:
            return _logo()

        if check_is_url(favicon_path):
            return flask.redirect(favicon_path)

        return send_from_dist(favicon_path, dist_folder=Settings.root_path)

    @bp.get("/_assets/logo")
    @cache.assets()
    def _logo():
        logo_path = controller.get_workspace().logo_url
        if not logo_path:
            return flask.abort(404)

        if check_is_url(logo_path):
            return flask.redirect(logo_path)

        return send_from_dist(logo_path, dist_folder=Settings.root_path)

    @bp.get("/_assets/background")
    @cache.assets()
    def _background():
        background_path = controller.get_workspace().theme

        if not background_path:
            return flask.abort(404)

        if check_is_url(background_path):
            return flask.redirect(background_path)

        return send_from_dist(background_path, dist_folder=Settings.root_path)

    @bp.route("/_hooks/<path:path>", methods=["POST", "GET", "PUT", "DELETE", "PATCH"])
    def hook_runner(path):
        hook = controller.get_hook_by_path(path)

        if not hook:
            flask.abort(404)

        if not hook.file:
            flask.abort(500)

        context = HookContext(
            request=extract_flask_request(flask.request),
            response=Response(headers={}, status=200, body=""),
        )

        connection = controller.repositories.producer.enqueue(hook.id, context)

        drain_until_response(connection)  # ExecutionStartedMessage

        try:
            response = normalize_response(drain_until_response(connection))

            if not response:
                flask.abort(500)
        finally:
            connection.close()

        return flask.Response(
            status=response.status,
            response=response.body,
            headers=response.headers,
        )

    def _run_page(path):
        page = controller.get_page_stage_by_path(path)

        if not page:
            flask.abort(404)

        if not page.file:
            flask.abort(500)

        # POST requests from function calls include the parent page execution ID
        page_execution_id = flask.request.headers.get("X-Page-Execution-Id")

        # Extract user JWT: Authorization header first (from page function calls),
        # fall back to editor_auth cookie (from headless browser / web editor)
        auth_header = flask.request.headers.get("Authorization", "")
        if auth_header.startswith("Bearer "):
            user_jwt = auth_header[7:]
        else:
            user_jwt = flask.request.cookies.get("editor_auth")

        context = PageContext(
            request=extract_flask_request(flask.request),
            response=Response(headers={}, status=200, body=""),
            page_path=path,
            page_execution_id=page_execution_id,
        )

        connection = controller.repositories.producer.enqueue(
            page.id, context, user_jwt=user_jwt
        )

        # First drain gets execution:started message
        start_msg = drain_until_response(connection)
        if not start_msg:
            connection.close()
            flask.abort(500)

        execution_id = None
        if isinstance(start_msg, dict) and start_msg.get("type") == "execution:started":
            execution_id = start_msg.get("executionId")

        msg = drain_until_response(connection)

        if not msg:
            connection.close()
            flask.abort(500)

        # Streaming response (generator functions)
        if isinstance(msg, dict) and msg.get("__page_stream__") == "start":

            def generate():
                try:
                    while True:
                        chunk = connection.recv()
                        if not isinstance(chunk, dict):
                            break
                        if chunk.get("__page_stream__") == "chunk":
                            yield json.dumps({"data": chunk["data"]}) + "\n"
                        elif chunk.get("__page_stream__") == "error":
                            yield json.dumps({"error": chunk["error"]}) + "\n"
                            break
                        elif chunk.get("__page_stream__") == "end":
                            break
                except (EOFError, BrokenPipeError):
                    pass
                finally:
                    connection.close()

            resp = flask.Response(
                status=msg["status"],
                response=generate(),
                headers=msg["headers"],
            )
            if execution_id:
                resp.headers["X-Execution-Id"] = execution_id
            if not IS_PRODUCTION:
                resp.headers["X-Abstra-Debug"] = "true"
            return resp

        # Regular response
        connection.close()

        response = normalize_response(msg)
        if not response:
            flask.abort(500)

        resp = flask.Response(
            status=response.status,
            response=response.body,
            headers=response.headers,
        )
        if execution_id:
            resp.headers["X-Execution-Id"] = execution_id
        if not IS_PRODUCTION:
            resp.headers["X-Abstra-Debug"] = "true"
        return resp

    def _serve_page_static(filename: str):
        token = flask.request.args.get("token")
        if not token:
            flask.abort(403)

        try:
            payload = pyjwt.decode(
                token, key=CLOUD_API_PROD_SHARED_TOKEN, algorithms=["HS256"]
            )
        except (pyjwt.ExpiredSignatureError, pyjwt.InvalidTokenError):
            flask.abort(403)

        if payload.get("asset") != filename:
            flask.abort(403)

        file_path = (Settings.root_path / filename).resolve()
        if not file_path.is_relative_to(Settings.root_path) or not file_path.is_file():
            flask.abort(404)

        mimetypes.add_type("application/javascript", ".js")
        mimetypes.add_type("text/css", ".css")
        return flask.send_file(file_path)

    @bp.get("/_page-home/<path:filename>")
    def page_home_static(filename):
        return _serve_page_static(filename)

    @bp.get("/_page/<page_path>/<path:filename>")
    def page_static(page_path, filename):
        return _serve_page_static(filename)

    @bp.route("/_page-home", methods=["GET", "POST"])
    @player_usage
    def page_home_runner():
        return _run_page("")

    @bp.route("/_page/<path:path>", methods=["GET", "POST"])
    @guard.by(PathArgSelector("path"))
    @player_usage
    def page_runner(path):
        return _run_page(path)

    @bp.post("/_logs/<execution_id>")
    def browser_logs(execution_id: str):
        """Receive browser console logs and store them as execution logs."""
        from datetime import datetime

        from abstra_internals.controllers.execution.execution_stdio import (
            BroadcastController,
        )
        from abstra_internals.repositories.execution_logs import LogEntry

        if not flask.request.json:
            flask.abort(400)

        body = flask.request.json
        logs = body.get("logs", [])
        stage_id = body.get("stageId", "")

        short_id = execution_id.split("-")[0]
        # Use high sequence base so browser logs sort after Python logs
        base_seq = 1_000_000 + controller.execution_logs_repository.get_sequence()

        for i, entry in enumerate(logs):
            level = entry.get("level", "log")
            message = entry.get("message", "")
            event = "stderr" if level in ("error", "warn") else "stdout"
            text = f"[RUN {short_id}] [BROWSER] {message}"
            controller.execution_logs_repository.save(
                LogEntry(
                    execution_id=execution_id,
                    stage_id=stage_id,
                    created_at=datetime.now(),
                    event=event,
                    payload={"text": text},
                    sequence=base_seq + i,
                )
            )
            # Broadcast directly via WebSocket (don't rely on file watcher)
            BroadcastController.broadcast(
                msg=serialize(
                    dict(
                        type="stdio",
                        payload=dict(
                            type=event,
                            log=text,
                            execution_id=execution_id,
                            stage_id=stage_id,
                        ),
                    )
                )
            )

        return {"ok": True}

    @bp.get("/_jobs")
    def list_jobs():
        if flask.request.headers.get("Shared-Token") != CLOUD_API_PROD_SHARED_TOKEN:
            flask.abort(401)

        # The scheduler needs all the jobs, including disabled ones, to schedule them.
        # The scheduler will always send the request to the lib to run the jobs, and the lib will check if the job is enabled or not.
        jobs = controller.get_jobs(include_disabled_jobs=True)

        # used by Scheduler Container - DO NOT CHANGE CONTRACT
        return [{"id": job.id, "schedule": job.schedule} for job in jobs]

    @bp.post("/_jobs/<path:id>")
    def job_runner(id):
        if flask.request.headers.get("Shared-Token") != CLOUD_API_PROD_SHARED_TOKEN:
            flask.abort(401)

        status = controller.get_job_status(id)
        if status == "not_found":
            flask.abort(404)

        if status == "disabled":
            return {"status": "disabled"}

        conn = controller.repositories.producer.enqueue(id, context=JobContext())
        try:
            conn.recv()
        finally:
            conn.close()

        return {"status": "running"}

    @bp.get("/")
    @cache.statics()
    def index():
        res = send_from_dist("player.html", "player.html")
        return res

    @bp.get("/<path:filename>")
    @cache.statics()
    def spa(filename: str):
        res = send_from_dist(filename, "player.html")
        return res

    setup_hook = Path("__setup__.py")
    if setup_hook.exists():
        module = __import__(path2module(setup_hook))
        if hasattr(module, "setup"):
            module.setup(bp)
        else:
            print(f"Could not find setup function in {setup_hook}")

    return bp



from abstra_internals.agents.tools.browser import _prepare_script


class TestPrepareScript:
    def test_simple_expression_unchanged(self):
        assert _prepare_script("window.location.href") == "window.location.href"

    def test_bare_return_wrapped_in_iife(self):
        result = _prepare_script("return window.location.href;")
        assert result == "(async () => { return window.location.href; })()"

    def test_multiline_with_return_wrapped(self):
        script = "const x = 1;\nreturn x;"
        result = _prepare_script(script)
        assert result == f"(async () => {{ {script} }})()"

    def test_expression_without_return_unchanged(self):
        assert (
            _prepare_script("document.querySelectorAll('a').length")
            == "document.querySelectorAll('a').length"
        )

    def test_return_inside_function_not_matched(self):
        script = "(function() { return 42; })()"
        assert _prepare_script(script) == script

    def test_return_at_start_of_line_after_statements(self):
        script = "let a = 1;\n  return a + 1;"
        result = _prepare_script(script)
        assert result.startswith("(async () => {")



from unittest import TestCase

from abstra_internals.entities.forms.widgets.library import Row, TextInput

from .template import TemplateRenderer


class TemplateRendererParseStateTest(TestCase):
    def test_simple(self):
        renderer = TemplateRenderer(
            [
                TextInput("name"),
                TextInput("age"),
            ]
        )

        raw_state = {"name": "Alice", "age": 30}
        state = renderer.parse_state(raw_state=raw_state, include_missing=True)

        self.assertEqual(state["name"], "Alice")
        self.assertEqual(state["age"], 30)

    def test_with_row(self):
        row = Row(
            [
                TextInput("first_name"),
                TextInput("last_name"),
            ]
        )
        renderer = TemplateRenderer(
            [
                TextInput("title"),
                # A Row widget containing two TextInput widgets
                row,
            ]
        )

        raw_state = {
            "title": "Ms.",
            row.key: {"first_name": "Alice", "last_name": "Smith"},
        }
        state = renderer.parse_state(raw_state=raw_state, include_missing=True)
        self.assertEqual(state["title"], "Ms.")
        self.assertEqual(state["first_name"], "Alice")
        self.assertEqual(state["last_name"], "Smith")



from abstra_internals.repositories.code_markers.providers.syntax import (
    SyntaxMarkerProvider,
)


class TestSyntaxMarkerProvider:
    def setup_method(self):
        self.provider = SyntaxMarkerProvider()

    def test_name(self):
        assert self.provider.name == "syntax"

    def test_supports_python(self):
        assert self.provider.supports_file_type("python") is True

    def test_does_not_support_json(self):
        assert self.provider.supports_file_type("json") is False

    def test_valid_code_returns_no_markers(self):
        code = """
def hello_world():
    print("Hello, world!")
    return "success"
"""
        markers = self.provider.get_markers(code)
        assert len(markers) == 0

    def test_syntax_error_returns_marker(self):
        code = "if True"  # Missing colon - reliable syntax error
        markers = self.provider.get_markers(code)
        assert len(markers) > 0
        assert markers[0].severity == "error"
        assert markers[0].source == "syntax"

    def test_marker_has_correct_structure(self):
        code = "if True"  # Missing colon
        markers = self.provider.get_markers(code)
        assert len(markers) > 0
        marker = markers[0]
        assert hasattr(marker, "line")
        assert hasattr(marker, "column")
        assert hasattr(marker, "until_line")
        assert hasattr(marker, "until_column")
        assert hasattr(marker, "message")
        assert hasattr(marker, "severity")
        assert hasattr(marker, "source")

    def test_marker_to_dict(self):
        code = "if True"  # Missing colon
        markers = self.provider.get_markers(code)
        assert len(markers) > 0
        marker_dict = markers[0].to_dict()
        assert "line" in marker_dict
        assert "column" in marker_dict
        assert "until_line" in marker_dict
        assert "until_column" in marker_dict
        assert "message" in marker_dict
        assert "severity" in marker_dict
        assert "source" in marker_dict
        assert marker_dict["source"] == "syntax"



import unittest
from unittest.mock import MagicMock

from abstra_internals.settings import Settings

Settings.set_root_path("/tmp")


class TestPlayerStringResponseDoesNotLeakInternals(unittest.TestCase):
    """Tests that unparseable string responses don't leak internal info."""

    def _setup_app_and_mock(self, recv_sequence):
        from abstra_internals.server.apps import get_cloud_app

        controller = MagicMock()

        page = MagicMock()
        page.id = "test-page-id"
        page.path = "test-page"
        page.file = "test_page.py"
        controller.get_page_stage_by_path = MagicMock(return_value=page)

        mock_conn = MagicMock()
        mock_conn.recv = MagicMock(side_effect=recv_sequence)
        mock_conn.poll = MagicMock(return_value=True)
        controller.repositories.producer.enqueue = MagicMock(return_value=mock_conn)

        app = get_cloud_app(controller)
        client = app.test_client()
        return client, mock_conn

    def test_traceback_string_not_leaked_to_client(self):
        """When response is an unparseable string (e.g. a traceback),
        the body should NOT contain that string — it would leak internals."""
        recv_sequence = [
            '{"type":"execution:started"}',
            "Traceback (most recent call last):\n  File 'worker.py', line 42\nKeyError: 'secret_key'",
        ]
        client, _ = self._setup_app_and_mock(recv_sequence)

        res = client.get("/_page/test-page")

        self.assertEqual(res.status_code, 500)
        self.assertNotIn(b"Traceback", res.data)
        self.assertNotIn(b"secret_key", res.data)

    def test_arbitrary_string_not_leaked(self):
        """Any arbitrary string should not be sent as-is to the client."""
        recv_sequence = [
            '{"type":"execution:started"}',
            "internal error details: password=hunter2",
        ]
        client, _ = self._setup_app_and_mock(recv_sequence)

        res = client.get("/_page/test-page")

        self.assertEqual(res.status_code, 500)
        self.assertNotIn(b"hunter2", res.data)


if __name__ == "__main__":
    unittest.main()



from tests.fixtures import BaseTest


class TestWorkspaceVersionApi(BaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.client = self.get_editor_flask_client()

    def test_get_version(self):
        """Test that the version endpoint returns the correct structure"""
        response = self.client.get("/_editor/api/workspace/version")
        self.assertEqual(response.status_code, 200)

        data = response.get_json()
        self.assertIn("version", data)
        self.assertIsInstance(data["version"], str)
        # Version should not be empty
        self.assertGreater(len(data["version"]), 0)

    def test_version_format(self):
        """Test that the version has expected format"""
        response = self.client.get("/_editor/api/workspace/version")
        data = response.get_json()

        version = data["version"]
        # Should have version format like x.y.z or x.y.z.dev or similar
        self.assertTrue(
            any(char.isdigit() for char in version),
            f"Version '{version}' should contain at least one digit",
        )



import mimetypes
import pkgutil
from datetime import datetime
from os.path import sep
from pathlib import Path
from typing import List, Optional, Union, cast

import flask

from abstra_internals.contracts_generated import (
    AbstraLibApiEditorCodebaseDirPostResponse,
    AbstraLibApiEditorCodebaseFilesDeleteResponse,
    AbstraLibApiEditorCodebaseFilesGetResponse,
    AbstraLibApiEditorCodebaseFilesGetResponseItem,
    AbstraLibApiEditorCodebaseFilesGetResponseItemStagesItem,
    AbstraLibApiEditorCodebaseFilesPatchResponse,
    AbstraLibApiEditorCodebaseFilesPutRequest,
    AbstraLibApiEditorCodebaseFilesPutResponse,
    AbstraLibApiEditorCodebaseSettingsGetResponse,
    CommonFileNode,
    CommonFileNodeType,
)
from abstra_internals.repositories.factory import Repositories
from abstra_internals.services.fs import FileSystemService
from abstra_internals.settings import Settings
from abstra_internals.templates import (
    new_agent_code,
    new_form_code,
    new_hook_code,
    new_job_code,
    new_page_code,
    new_script_code,
)
from abstra_internals.utils.file import safe_write_file


class CodebaseController:
    def __init__(self, repos: Repositories):
        self.repos = repos

    def file_node(
        self, path: Path, type_override: Optional[str] = None
    ) -> CommonFileNode:
        if type_override:
            is_dir = type_override == "package"
            type_name = type_override
            size = 0
            mtime = 0.0
        else:
            is_dir = path.is_dir()
            type_name = "directory" if is_dir else "file"
            try:
                stats = path.stat()
                size = stats.st_size
                mtime = stats.st_mtime
            except OSError:
                size = 0
                mtime = 0.0

        return CommonFileNode(
            path_parts=list(path.parts),
            size=size,
            last_modified=datetime.fromtimestamp(mtime),
            type=cast(CommonFileNodeType, type_name),
        )

    def list_files(
        self, path: Union[str, Path, None], mode: str = "file"
    ) -> AbstraLibApiEditorCodebaseFilesGetResponse:
        if path is None:
            path = Path()
        elif isinstance(path, str):
            path = Path(path)
        elif not isinstance(path, Path):
            raise ValueError(f"Invalid path: {path}")

        project = self.repos.project.load(include_disabled_stages=True)

        if mode == "module":
            modules = []
            search_path = (
                str(Settings.root_path / path)
                if str(path) != "."
                else str(Settings.root_path)
            )
            for _, name, ispkg in pkgutil.iter_modules([search_path]):
                # Basic check to avoid unrelated modules if path is not in sys.path
                # But pkgutil.iter_modules needs a list of paths
                module_path = path / name
                modules.append(
                    AbstraLibApiEditorCodebaseFilesGetResponseItem(
                        file=self.file_node(
                            module_path, type_override="package" if ispkg else "module"
                        ),
                        stages=[],
                    )
                )
            return modules

        allowed_suffixes = None
        if mode == "image":
            allowed_suffixes = [
                ".png",
                ".jpg",
                ".jpeg",
                ".gif",
                ".svg",
                ".webp",
                ".jfif",
                ".pjp",
                ".pjpeg",
            ]
        elif mode == "python-file":
            allowed_suffixes = [".py"]

        # Ensure we are listing relative to root
        full_path = Settings.root_path / path

        ALWAYS_HIDDEN = {
            ".git",
            ".venv",
            "venv",
            "__pycache__",
            ".pyrefly_buffer.py",
            "pyrefly.toml",
        }

        return [
            AbstraLibApiEditorCodebaseFilesGetResponseItem(
                file=self.file_node(
                    child_path.relative_to(Settings.root_path)
                    if child_path.is_relative_to(Settings.root_path)
                    else child_path
                ),
                stages=[
                    AbstraLibApiEditorCodebaseFilesGetResponseItemStagesItem(
                        id=stage.id,
                        type=stage.type_name,
                    )
                    for stage in project.get_stages_by_file_path(child_path)
                    if stage.type_name != "component"
                ],
            )
            for child_path in FileSystemService.list_files(
                full_path,
                include_dirs=True,
                use_ignore=False,
                recursive=False,
                allowed_suffixes=allowed_suffixes,
            )
            if child_path != full_path and child_path.name not in ALWAYS_HIDDEN
        ]

    def init_file(self, path: str, type: str):
        if type == "scripts":
            code = new_script_code
        elif type == "forms":
            code = new_form_code
        elif type == "hooks":
            code = new_hook_code
        elif type == "jobs":
            code = new_job_code
        elif type == "pages":
            code = new_page_code
        elif type == "agents":
            code = new_agent_code
        else:
            raise ValueError(f"Invalid type: {type}")

        self.create_file(path, code.encode("utf-8"), overwrite=False)

    def create_file(
        self, path, content: Optional[bytes] = None, overwrite: bool = False
    ) -> CommonFileNode:
        from abstra_internals.settings import Settings

        if isinstance(path, str):
            path = Path(path)
        elif not isinstance(path, Path):
            raise ValueError(f"Invalid path: {path}")

        relative_path = path

        if not path.is_absolute():
            path = Settings.root_path / path

        if path.exists() and not overwrite:
            raise FileExistsError(f"File already exists: {path}")

        if path.exists() and overwrite:
            path.unlink()

        path.parent.mkdir(parents=True, exist_ok=True)

        if content is not None:
            path.write_bytes(content)
        else:
            path.touch()
        return CommonFileNode(
            path_parts=list(relative_path.parts),
            size=path.stat().st_size,
            last_modified=datetime.fromtimestamp(path.stat().st_mtime),
            type="directory" if path.is_dir() else "file",
        )

    def delete_file(
        self, path_parts: List[str]
    ) -> AbstraLibApiEditorCodebaseFilesDeleteResponse:
        from abstra_internals.settings import Settings

        path = Path(*path_parts)
        if not path.is_absolute():
            path = Settings.root_path / path

        if path.is_dir():
            FileSystemService.rm_tree(path)
        else:
            path.unlink()
        return AbstraLibApiEditorCodebaseFilesDeleteResponse(ok=True)

    def rename_file(
        self, path, new_name
    ) -> AbstraLibApiEditorCodebaseFilesPatchResponse:
        from abstra_internals.settings import Settings

        if isinstance(path, str):
            path = Path(path)
        elif isinstance(path, List):
            path = Path(*path)
        elif not isinstance(path, Path):
            raise ValueError(f"Invalid path: {path}")

        if not path.is_absolute():
            path = Settings.root_path / path

        if isinstance(new_name, str):
            new_path = path.parent / new_name
        elif isinstance(new_name, List):
            new_path = path.parent / Path(*new_name)
        else:
            raise ValueError(f"Invalid new name: {new_name}")

        # Check if the renamed file is a workflow stage
        project = self.repos.project.load(include_disabled_stages=True)
        stages = project.get_stages_by_file_path(path)

        path.rename(new_path)

        if stages:
            from abstra_internals.settings import Settings

            relative_new_path = new_path.relative_to(Settings.root_path)
            stages[0].update({"file": str(relative_new_path)})
            self.repos.project.save(project)

        return AbstraLibApiEditorCodebaseFilesPatchResponse(ok=True)

    def edit_file(
        self, path, content: AbstraLibApiEditorCodebaseFilesPutRequest
    ) -> AbstraLibApiEditorCodebaseFilesPutResponse:
        if isinstance(path, str):
            path = Path(path)
        elif not isinstance(path, Path):
            raise ValueError(f"Invalid path: {path}")

        if not path.is_absolute():
            path = Settings.root_path / path

        resolved_path = path.resolve()
        if not resolved_path.is_relative_to(Settings.root_path.resolve()):
            raise ValueError(f"Path is outside project root: {path}")

        write_ok = safe_write_file(resolved_path, content.content)
        if not write_ok:
            return AbstraLibApiEditorCodebaseFilesPutResponse(ok=False)

        return AbstraLibApiEditorCodebaseFilesPutResponse(ok=True)

    def get_file(self, path):
        if isinstance(path, str):
            path = Path(path)
        elif not isinstance(path, Path):
            raise ValueError(f"Invalid path: {path}")
        resolved = (
            (Settings.root_path / path).resolve()
            if not path.is_absolute()
            else path.resolve()
        )
        if not path.is_absolute() and not resolved.is_relative_to(
            Settings.root_path.resolve()
        ):
            flask.abort(403)
        if not resolved.exists():
            flask.abort(404)
        mtype, _ = mimetypes.guess_type(resolved)
        return flask.send_file(resolved, mimetype=mtype)

    def mkdir(
        self, path: Union[str, Path, List[str]]
    ) -> AbstraLibApiEditorCodebaseDirPostResponse:
        from abstra_internals.settings import Settings

        if isinstance(path, str):
            path = Path(path)
        elif isinstance(path, list):
            path = Path(*path)
        elif not isinstance(path, Path):
            raise ValueError(f"Invalid path: {path}")

        if not path.is_absolute():
            path = Settings.root_path / path

        path.mkdir(parents=True, exist_ok=True)
        return AbstraLibApiEditorCodebaseDirPostResponse(ok=True)

    def check_file(self, path: str) -> dict:
        full_path = Settings.root_path / path
        return {"exists": full_path.exists()}

    def check_files(self, paths: List[str]) -> dict:
        return {path: (Settings.root_path / path).exists() for path in paths}

    def settings(self) -> AbstraLibApiEditorCodebaseSettingsGetResponse:
        return AbstraLibApiEditorCodebaseSettingsGetResponse(
            separator=sep,
        )



import unittest
from unittest.mock import MagicMock, patch

from abstra_internals.controllers.main import MainController
from abstra_internals.entities.execution_context import (
    Request,
    Response,
)
from abstra_internals.settings import Settings

Settings.set_root_path("/tmp")


def _make_controller_and_mock_connection(recv_sequence):
    """Create a MainController with mocked producer and connection.

    recv_sequence: list of values that connection.recv() will return in order.
    """
    controller = MagicMock()

    # Bind the real method so we can test it
    controller.run_page_stage = MainController.run_page_stage.__get__(controller)

    page = MagicMock()
    page.id = "test-page-id"
    page.path = "test-page"
    controller.get_page_stage = MagicMock(return_value=page)

    mock_conn = MagicMock()
    mock_conn.recv = MagicMock(side_effect=recv_sequence)
    mock_conn.poll = MagicMock(return_value=True)
    controller.repositories.producer.enqueue = MagicMock(return_value=mock_conn)

    return controller, mock_conn


class TestRunPageStageDrainsStdioMessages(unittest.TestCase):
    """Tests that run_page_stage correctly drains intermediate stdio messages
    from the RabbitMQ session queue and finds the actual page Response."""

    def test_normal_dict_response(self):
        """When the second recv() returns a proper response dict, it works."""
        recv_sequence = [
            '{"type":"execution:started","data":{"executionId":"abc"}}',
            {
                "headers": {"Content-Type": "text/html"},
                "status": 200,
                "body": "<h1>OK</h1>",
            },
        ]
        controller, mock_conn = _make_controller_and_mock_connection(recv_sequence)
        request = Request(headers={}, body="", query_params={}, method="GET")

        result = controller.run_page_stage("test-page-id", request)

        self.assertEqual(result["status"], 200)
        self.assertEqual(result["body"], "<h1>OK</h1>")
        mock_conn.close.assert_called_once()

    def test_string_response_is_deserialized(self):
        """When the response arrives as a JSON string (text/plain), it is deserialized."""
        recv_sequence = [
            '{"type":"execution:started"}',
            '{"headers": {}, "status": 200, "body": "hello"}',
        ]
        controller, mock_conn = _make_controller_and_mock_connection(recv_sequence)
        request = Request(headers={}, body="", query_params={}, method="GET")

        result = controller.run_page_stage("test-page-id", request)

        self.assertEqual(result["status"], 200)
        self.assertEqual(result["body"], "hello")

    def test_drains_single_stdio_batch_before_response(self):
        """A stdio_batch between execution:started and Response is skipped."""
        recv_sequence = [
            '{"type":"execution:started"}',
            {"type": "stdio_batch", "payload": [{"type": "stdout", "log": "hello"}]},
            {"headers": {}, "status": 200, "body": "<h1>Page</h1>"},
        ]
        controller, mock_conn = _make_controller_and_mock_connection(recv_sequence)
        request = Request(headers={}, body="", query_params={}, method="GET")

        result = controller.run_page_stage("test-page-id", request)

        self.assertEqual(result["status"], 200)
        self.assertEqual(result["body"], "<h1>Page</h1>")

    def test_drains_multiple_stdio_batches_before_response(self):
        """Multiple stdio_batch messages are all drained before the Response."""
        recv_sequence = [
            '{"type":"execution:started"}',
            {"type": "stdio_batch", "payload": [{"type": "stdout", "log": "line1"}]},
            {"type": "stdio_batch", "payload": [{"type": "stdout", "log": "line2"}]},
            {"type": "stdio", "payload": {"type": "stderr", "log": "warn"}},
            {
                "headers": {"Content-Type": "text/html"},
                "status": 200,
                "body": "<h1>OK</h1>",
            },
        ]
        controller, mock_conn = _make_controller_and_mock_connection(recv_sequence)
        request = Request(headers={}, body="", query_params={}, method="GET")

        result = controller.run_page_stage("test-page-id", request)

        self.assertEqual(result["status"], 200)
        self.assertEqual(result["body"], "<h1>OK</h1>")

    def test_response_object_returned_directly(self):
        """When connection.recv() returns a Response instance, it's used directly."""
        response_obj = Response(headers={"X-Custom": "val"}, status=201, body="created")
        recv_sequence = [
            '{"type":"execution:started"}',
            response_obj,
        ]
        controller, mock_conn = _make_controller_and_mock_connection(recv_sequence)
        request = Request(headers={}, body="", query_params={}, method="GET")

        result = controller.run_page_stage("test-page-id", request)

        self.assertEqual(result["status"], 201)
        self.assertEqual(result["body"], "created")
        self.assertEqual(result["headers"], {"X-Custom": "val"})

    def test_execution_ended_string_without_response(self):
        """When execution:ended arrives without a Response (e.g. worker crashed),
        it's parsed as a dict and returns 500 — not a valid page response."""
        recv_sequence = [
            '{"type":"execution:started"}',
            '{"type":"execution:ended","data":{"exitStatus":"EXCEPTION"}}',
        ]
        controller, mock_conn = _make_controller_and_mock_connection(recv_sequence)
        request = Request(headers={}, body="", query_params={}, method="GET")

        result = controller.run_page_stage("test-page-id", request)

        self.assertEqual(result["status"], 500)
        self.assertEqual(result["body"], "Internal Server Error")

    @patch("abstra_internals.controllers.main.flask")
    def test_none_response_aborts_500(self, mock_flask):
        """When connection.recv() returns None, the server aborts with 500."""
        mock_flask.abort = MagicMock(side_effect=Exception("abort"))
        recv_sequence = [
            '{"type":"execution:started"}',
            None,
        ]
        controller, mock_conn = _make_controller_and_mock_connection(recv_sequence)
        request = Request(headers={}, body="", query_params={}, method="GET")

        with self.assertRaises(Exception):
            controller.run_page_stage("test-page-id", request)

        mock_flask.abort.assert_called_once_with(500)


class TestRunPageStageStartMsgValidation(unittest.TestCase):
    """Tests that run_page_stage fails fast when the start message is invalid."""

    @patch("abstra_internals.controllers.main.flask")
    def test_none_start_msg_aborts_500_immediately(self, mock_flask):
        """If first drain returns None (timeout), abort 500 without waiting
        for a second drain that would waste another 30s."""
        mock_flask.abort = MagicMock(side_effect=Exception("abort"))
        controller, mock_conn = _make_controller_and_mock_connection([None])
        request = Request(headers={}, body="", query_params={}, method="GET")

        with self.assertRaises(Exception):
            controller.run_page_stage("test-page-id", request)

        mock_flask.abort.assert_called_once_with(500)
        mock_conn.close.assert_called_once()



"""
Git types and data structures shared across implementations.
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple


@dataclass
class GitCommit:
    hash: str
    message: str
    author: str
    date: str


@dataclass
class AheadBehindInfo:
    ahead: int = 0
    behind: int = 0


@dataclass
class RemoteGitStatus:
    success: bool
    remote_commit: Optional[str] = None
    branch: str = "main"
    has_commits: bool = False
    available_branches: List[str] = field(default_factory=list)
    message: Optional[str] = None
    ahead_behind: Optional[AheadBehindInfo] = None
    has_remote_changes: Optional[bool] = None
    has_conflicts: Optional[bool] = None


@dataclass
class GitFileChange:
    path: str
    status: str  # added, deleted, modified, renamed, untracked
    status_code: str  # Git porcelain status code (e.g., "M ", "A ", "??")


@dataclass
class LargeFileInfo:
    path: str
    size_bytes: int
    size_human: str


@dataclass
class CleanupResult:
    success: bool
    directories_cleaned: int = 0
    bytes_freed: int = 0


@dataclass
class MaintenanceResult:
    success: bool
    cleanup: Optional["CleanupResult"] = None
    gc_ran: bool = False


@dataclass
class GitStatusResponse:
    available: bool
    git_installed: bool
    is_monorepo: bool
    branch: str
    branches: List[str]
    last_commit: Optional[GitCommit]
    has_changes: bool
    changed_files: List[str]
    changed_files_with_status: List[GitFileChange]
    remote_status: Optional[RemoteGitStatus] = None

    def to_dict(self) -> Dict[str, Any]:
        result = {
            "available": self.available,
            "gitInstalled": self.git_installed,
            "isMonorepo": self.is_monorepo,
            "branch": self.branch,
            "branches": self.branches,
            "lastCommit": {
                "hash": self.last_commit.hash,
                "message": self.last_commit.message,
                "author": self.last_commit.author,
                "date": self.last_commit.date,
            }
            if self.last_commit
            else None,
            "hasChanges": self.has_changes,
            "changedFiles": self.changed_files,
            "changedFilesWithStatus": [
                {
                    "path": file_change.path,
                    "status": file_change.status,
                    "statusCode": file_change.status_code,
                }
                for file_change in self.changed_files_with_status
            ],
        }

        if self.remote_status:
            result["remoteStatus"] = {
                "success": self.remote_status.success,
                "remoteCommit": self.remote_status.remote_commit,
                "hasRemoteChanges": self.remote_status.has_remote_changes or False,
                "hasConflicts": self.remote_status.has_conflicts or False,
                "message": self.remote_status.message,
                "availableBranches": self.remote_status.available_branches,
                "aheadBehind": {
                    "ahead": self.remote_status.ahead_behind.ahead,
                    "behind": self.remote_status.ahead_behind.behind,
                }
                if self.remote_status.ahead_behind
                else None,
            }
        else:
            result["remoteStatus"] = None

        return result


@dataclass
class GitStatus:
    available: bool
    git_installed: bool = True
    is_monorepo: bool = False
    branch: str = ""
    branches: List[str] = field(default_factory=list)
    last_commit: Optional[GitCommit] = None
    has_changes: bool = False
    changed_files: List[str] = field(default_factory=list)
    changed_files_with_status: List[GitFileChange] = field(default_factory=list)
    remote_status: Optional[RemoteGitStatus] = None


class GitRepositoryInterface(ABC):
    """Abstract interface for Git repository implementations"""

    def __init__(self, working_directory: Path):
        self.working_directory = working_directory

    @abstractmethod
    def is_git_available(self) -> bool:
        """Check if git functionality is available"""
        pass

    @abstractmethod
    def is_git_repository(self) -> bool:
        """Check if current directory is a git repository"""
        pass

    @abstractmethod
    def find_git_root(self, start_path: Optional[Path] = None) -> Optional[Path]:
        """Find the root directory of the git repository"""
        pass

    @abstractmethod
    def configure_git_user(self, fallback_email: str, fallback_name: str):
        """Ensure git user is configured for commits (required in CI environments)"""
        pass

    @abstractmethod
    def get_current_branch(self) -> Optional[str]:
        """Get the current branch name"""
        pass

    @abstractmethod
    def get_all_branches(self) -> List[str]:
        """Get all local branches"""
        pass

    @abstractmethod
    def get_last_commit(self) -> Optional[GitCommit]:
        """Get information about the last commit"""
        pass

    @abstractmethod
    def get_commit_history(
        self, limit: int = 10, offset: int = 0, branch: Optional[str] = None
    ) -> List[GitCommit]:
        """Get commit history with pagination"""
        pass

    @abstractmethod
    def get_changed_files(self) -> List[str]:
        """Get list of changed files"""
        pass

    @abstractmethod
    def get_changed_files_with_status(self) -> List[GitFileChange]:
        """Get list of changed files with their git status"""
        pass

    @abstractmethod
    def has_uncommitted_changes(self) -> bool:
        """Check if there are uncommitted changes"""
        pass

    @abstractmethod
    def init_repository(self) -> bool:
        """Initialize a new git repository"""
        pass

    @abstractmethod
    def get_repository_status(self) -> GitStatus:
        """Get comprehensive repository status"""
        pass

    @abstractmethod
    def checkout_branch(self, branch_name: str) -> Tuple[bool, Optional[str]]:
        """Switch to a different branch"""
        pass

    @abstractmethod
    def checkout_commit(self, commit_hash: str) -> Tuple[bool, Optional[str]]:
        """Switch to a specific commit (detached HEAD state)"""
        pass

    @abstractmethod
    def pull_changes(
        self,
        strategy: str = "merge",
        allow_unrelated: bool = True,
        conflict_resolution: Optional[str] = "theirs",
    ) -> Tuple[bool, Optional[str]]:
        """Pull changes from remote repository"""
        pass

    @abstractmethod
    def commit_changes(
        self, message: str, author: Optional[str] = None
    ) -> Tuple[bool, Optional[str]]:
        """Commit changes with a message"""
        pass

    @abstractmethod
    def stash_changes(self, message: str = "WIP") -> Tuple[bool, Optional[str]]:
        """Stash uncommitted changes"""
        pass

    @abstractmethod
    def get_remotes(self) -> List[str]:
        """Get list of remote names"""
        pass

    @abstractmethod
    def has_remote(self, remote_name: str) -> bool:
        """Check if a remote exists"""
        pass

    @abstractmethod
    def add_remote(self, remote_name: str, remote_url: str) -> bool:
        """Add a remote to the repository"""
        pass

    @abstractmethod
    def set_remote_url(self, remote_name: str, remote_url: str) -> bool:
        """Set/update the URL for a remote"""
        pass

    @abstractmethod
    def set_git_config(self, key: str, value: str, local: bool = True) -> bool:
        """Set a git configuration value"""
        pass

    @abstractmethod
    def get_git_config(self, key: str, local: bool = True) -> Optional[str]:
        """Get a git configuration value"""
        pass

    @abstractmethod
    def push_and_deploy(self, branch: str = "main") -> Tuple[bool, Optional[str]]:
        """Deploy to Abstra remote (push to abstra remote)"""
        pass

    @abstractmethod
    def revert_commit(self, commit_hash: str) -> Tuple[bool, Optional[str]]:
        """Restore content from a previous commit by creating a new commit with that content"""
        pass

    @abstractmethod
    def check_merge_conflicts(self, remote_commit: str) -> bool:
        """Check if merging with remote commit would cause conflicts"""
        pass

    @abstractmethod
    def get_ahead_behind_count(
        self, local_commit: str, remote_commit: str
    ) -> Tuple[int, int]:
        """Calculate ahead/behind count between local and remote commits"""
        pass

    @abstractmethod
    def check_ignore(self, path: Path) -> bool:
        """Check if a path should be ignored according to .gitignore rules"""
        pass

    @abstractmethod
    def check_ignore_batch(self, paths: List[Path]) -> Dict[Path, bool]:
        """Check multiple paths at once for better performance"""
        pass

    @abstractmethod
    def untrack_path(self, path: Path):
        """Untrack a path from git (git rm --cached)"""
        pass

    @abstractmethod
    def get_large_files(
        self, max_size_bytes: int = 5 * 1024 * 1024
    ) -> List["LargeFileInfo"]:
        """Get list of changed files that exceed the size limit (default 5MB)"""
        pass

    @abstractmethod
    def add_to_gitignore(self, paths: List[str]) -> Tuple[bool, Optional[str]]:
        """Add paths to .gitignore file"""
        pass

    @abstractmethod
    def cleanup_orphan_temp_objects(self) -> "CleanupResult":
        """Clean up orphan temporary object directories from failed git operations."""
        pass

    @abstractmethod
    def run_maintenance(self) -> "MaintenanceResult":
        """Run git repository maintenance tasks."""
        pass



from unittest.mock import patch

from abstra_internals.repositories.linter.rules.invalid_package_in_requirements import (
    InvalidPackageInRequirements,
)
from abstra_internals.services.pypi_cache import PyPIVerificationCache
from abstra_internals.services.requirements import RequirementsRepository
from tests.fixtures import BaseTest


class TestInvalidPackageInRequirements(BaseTest):
    def setUp(self):
        super().setUp()
        # Clear the cache before each test
        PyPIVerificationCache.clear_cache()

    def test_default_valid(self):
        with patch.object(PyPIVerificationCache, "_query_pypi", return_value=True):
            issues = InvalidPackageInRequirements().find_issues()
            self.assertEqual(len(issues), 0)

    def test_valid_with_no_requirements_file(self):
        requirements_txt = self.root / "requirements.txt"
        requirements_txt.unlink()
        issues = InvalidPackageInRequirements().find_issues()
        self.assertEqual(len(issues), 0)

    def test_valid_with_existing_packages(self):
        requirements_txt = self.root / "requirements.txt"
        requirements_txt.write_text("requests==2.28.0\nflask==2.0.0")

        with patch.object(PyPIVerificationCache, "_query_pypi", return_value=True):
            issues = InvalidPackageInRequirements().find_issues()
            self.assertEqual(len(issues), 0)

    def test_invalid_with_nonexistent_package(self):
        requirements_txt = self.root / "requirements.txt"
        requirements_txt.write_text("nonexistent-package-xyz==1.0.0")

        with patch.object(PyPIVerificationCache, "_query_pypi", return_value=False):
            issues = InvalidPackageInRequirements().find_issues()
            self.assertEqual(len(issues), 1)
            self.assertIn("nonexistent-package-xyz", issues[0].label)

    def test_fix_removes_invalid_package(self):
        requirements_txt = self.root / "requirements.txt"
        requirements_txt.write_text("valid-package==1.0.0\ninvalid-package==1.0.0")

        def mock_query(package_name):
            return package_name != "invalid-package"

        with patch.object(PyPIVerificationCache, "_query_pypi", side_effect=mock_query):
            issues = InvalidPackageInRequirements().find_issues()
            self.assertEqual(len(issues), 1)

            # Apply the fix
            issues[0].fixes[0].fix()

            # Clear cache to re-check
            PyPIVerificationCache.clear_cache()

            # Verify the package was removed
            requirements = RequirementsRepository.load()
            package_names = [lib.name for lib in requirements.libraries]
            self.assertNotIn("invalid-package", package_names)
            self.assertIn("valid-package", package_names)

    def test_skips_packages_with_urls(self):
        requirements_txt = self.root / "requirements.txt"
        requirements_txt.write_text("mypackage @ https://example.com/mypackage.tar.gz")

        with patch.object(PyPIVerificationCache, "_query_pypi", return_value=False):
            issues = InvalidPackageInRequirements().find_issues()
            # Should not report issue for URL-based packages
            self.assertEqual(len(issues), 0)

    def test_cache_prevents_duplicate_queries(self):
        requirements_txt = self.root / "requirements.txt"
        requirements_txt.write_text("cached-package==1.0.0")

        query_count = 0

        def mock_query(package_name):
            nonlocal query_count
            query_count += 1
            return True

        with patch.object(PyPIVerificationCache, "_query_pypi", side_effect=mock_query):
            # First call should query PyPI
            InvalidPackageInRequirements().find_issues()
            self.assertEqual(query_count, 1)

            # Second call should use cache
            InvalidPackageInRequirements().find_issues()
            self.assertEqual(query_count, 1)  # Still 1, used cache



import webbrowser
from typing import List

from abstra_internals.environment import EDITOR_MODE, RABBITMQ_CONNECTION_URI
from abstra_internals.logger import AbstraLogger
from abstra_internals.repositories.linter.models import (
    LinterFix,
    LinterIssue,
    LinterRule,
)
from abstra_internals.repositories.producer import WebEditorControlProducerRepository
from abstra_internals.utils.platform import is_windows
from abstra_internals.version import PackageVersionManager, VersionStatus

RELEASE_NOTES_URL = "https://github.com/abstra-app/abstra-lib/releases"


def _update_lib_version():
    import os
    import subprocess
    import sys

    try:
        subprocess.check_call(
            [sys.executable, "-m", "pip", "install", "--upgrade", "abstra"]
        )

        if EDITOR_MODE == "web":
            # Restart workers first so they reload with the new code
            if RABBITMQ_CONNECTION_URI:
                try:
                    control_producer = WebEditorControlProducerRepository(
                        RABBITMQ_CONNECTION_URI
                    )
                    control_producer.restart_workers()
                    AbstraLogger.warning(
                        "[UpdateAbstra] Sent restart signal to workers"
                    )
                except Exception as e:
                    AbstraLogger.error(
                        f"[UpdateAbstra] Failed to send restart signal to workers: {e}"
                    )

            # Force immediate process termination for Kubernetes to restart
            # sys.exit(0) doesn't work because Flask runs in threaded mode
            # os._exit() bypasses threads and terminates the process directly
            AbstraLogger.warning("[UpdateAbstra] Restarting editor after update")
            os._exit(0)
        else:
            # Local: restart process in place
            os.execv(sys.executable, [sys.executable] + sys.argv)
    except Exception as e:
        AbstraLogger.error(f"[UpdateAbstra] Failed to update Abstra: {e}")


class UpdateAbstraVersion(LinterFix):
    def __init__(self) -> None:
        self.label = "Update Abstra Editor version"

    def fix(self):
        _update_lib_version()


class OpenChangeLog(LinterFix):
    def __init__(self) -> None:
        self.label = "Open the release notes"

    def fix(self):
        webbrowser.open(RELEASE_NOTES_URL)


class NewVersionOfAbstraAvailableFound(LinterIssue):
    def __init__(self) -> None:
        package_version = PackageVersionManager("abstra")
        self.label = f"Latest version is {package_version.cached_latest_version}, but you have {package_version.current_local_version}. Updating may take up to 2 minutes."
        if is_windows():
            self.fixes = [OpenChangeLog()]
        else:
            self.fixes = [UpdateAbstraVersion()]


class NewVersionOfAbstraAvailable(LinterRule):
    label = "A new version of Abstra Editor is available"
    type = "info"

    def find_issues(self) -> List[LinterIssue]:
        package_version = PackageVersionManager("abstra")
        version_status = package_version.get_version_status()
        is_there_a_new_version = version_status == VersionStatus.OUT_OF_DATE

        if is_there_a_new_version is True:
            return [NewVersionOfAbstraAvailableFound()]
        else:
            return []



from .base_migration import Migration


class Migration016(Migration):
    @staticmethod
    def target_version() -> str:
        return "16.0"

    @staticmethod
    def source_version() -> str:
        return "15.0"

    def _migrate(self) -> None:
        for form_data in self.data.get("forms", []):
            form_data.pop("allow_restart", None)
            form_data.pop("restart_button_text", None)
            form_data.pop("welcome_title", None)



import pathlib
import unittest

from abstra_internals.templates import new_hook_code
from tests.fixtures import BaseTest


class TestCRUDHooks(BaseTest):
    def setUp(self) -> None:
        super().setUp()
        self.client = self.get_editor_flask_client()

    def test_api_list(self):
        hooks = self.client.get("/_editor/api/hooks/").get_json()
        self.assertEqual(len(hooks), 0)
        self.client.post(
            "/_editor/api/hooks/", json={"title": "New hook", "file": "hook.py"}
        )
        hooks = self.client.get("/_editor/api/hooks/").get_json()
        self.assertEqual(len(hooks), 1)

    def test_change_path(self):
        hook = self.client.post(
            "/_editor/api/hooks/", json={"title": "New hook", "file": "hook.py"}
        ).get_json()
        self.client.put("/_editor/api/hooks/" + hook["id"], json={"path": "new_path"})
        hooks = self.client.get("/_editor/api/hooks/").get_json()
        self.assertEqual(len(hooks), 1)
        self.assertEqual(hooks[0]["path"], "new-path")

    def test_create_hook_with_right_template(self):
        hook = self.client.post(
            "/_editor/api/hooks/", json={"title": "New hook", "file": "hook.py"}
        ).get_json()
        file_content = pathlib.Path(hook["file"]).read_text(encoding="utf-8")
        self.assertEqual(file_content, new_hook_code)


if __name__ == "__main__":
    unittest.main()



import ast
import unittest

from abstra_internals.repositories.linter.rules.missing_render_in_page import (
    MissingRenderInPage,
)


class TestMissingRenderInPage(unittest.TestCase):
    def setUp(self):
        self.rule = MissingRenderInPage()

    def test_detects_render_function(self):
        code = """
from abstra.pages import register_function

@register_function
def __render__():
    return '<div>hello</div>'
"""
        tree = ast.parse(code)
        self.assertTrue(self.rule._has_render_function(tree))

    def test_detects_render_with_module_decorator(self):
        code = """
import abstra.pages as pages

@pages.register_function
def __render__():
    return '<div>hello</div>'
"""
        tree = ast.parse(code)
        self.assertTrue(self.rule._has_render_function(tree))

    def test_flags_missing_render(self):
        code = """
from abstra.pages import register_function

@register_function
def get_data():
    return {}
"""
        tree = ast.parse(code)
        self.assertFalse(self.rule._has_render_function(tree))

    def test_flags_render_without_decorator(self):
        code = """
def __render__():
    return '<div>hello</div>'
"""
        tree = ast.parse(code)
        self.assertFalse(self.rule._has_render_function(tree))

    def test_flags_empty_file(self):
        code = ""
        tree = ast.parse(code)
        self.assertFalse(self.rule._has_render_function(tree))

    def test_flags_wrong_function_name_with_decorator(self):
        code = """
from abstra.pages import register_function

@register_function
def render():
    return '<div>hello</div>'
"""
        tree = ast.parse(code)
        self.assertFalse(self.rule._has_render_function(tree))

    def test_detects_render_alongside_other_functions(self):
        code = """
from abstra.pages import register_function

@register_function
def get_data():
    return {}

@register_function
def __render__():
    return '<div>hello</div>'
"""
        tree = ast.parse(code)
        self.assertTrue(self.rule._has_render_function(tree))



from unittest import TestCase

from .geometry import Matrix2D, Point2D, Rect


class TestMatrix2D(TestCase):
    def test_inverse_identity(self):
        """Test inverse of identity matrix"""
        identity = Matrix2D.from_rows(
            [
                [1, 0],
                [0, 1],
            ]
        )
        inverse = identity.inverse()
        expected = Matrix2D.from_rows(
            [
                [1, 0],
                [0, 1],
            ]
        )
        for i, value in enumerate(inverse.data):
            self.assertAlmostEqual(value, expected.data[i], places=7)

    def test_determinant_identity(self):
        """Test determinant of identity matrix"""
        identity = Matrix2D.from_rows(
            [
                [1, 0],
                [0, 1],
            ]
        )
        det = identity.determinant()
        self.assertEqual(det, 1)


class TestPoint2D(TestCase):
    def test_cast_to_rect_from_top(self):
        """Test casting point to rectangle from the top"""
        p = Point2D(0, 100)
        rect = Rect(Point2D(0, 0), 10, 10)
        result = p.cast_to_rect(rect)
        self.assertEqual(result.x, 0)
        self.assertEqual(result.y, 5)

    def test_cast_to_rect_from_bottom(self):
        """Test casting point to rectangle from the bottom"""
        p = Point2D(0, -100)
        rect = Rect(Point2D(0, 0), 10, 10)
        result = p.cast_to_rect(rect)
        self.assertEqual(result.x, 0)
        self.assertEqual(result.y, -5)


class TestEdge(TestCase):
    def test_intersect_parallel_lines(self):
        """Test intersection of parallel lines"""
        line1 = Point2D(0, 0).line_to(Point2D(10, 0))
        line2 = Point2D(0, 1).line_to(Point2D(10, 1))
        intersection = line1.intersect(line2)
        self.assertIsNone(intersection)

    def test_intersect_crossing_lines(self):
        """Test intersection of crossing lines"""
        line1 = Point2D(0, 0).line_to(Point2D(10, 10))
        line2 = Point2D(0, 10).line_to(Point2D(10, 0))
        intersection = line1.intersect(line2)
        self.assertIsNotNone(intersection)
        if intersection:  # Type guard for None check
            self.assertEqual(intersection.x, 5)
            self.assertEqual(intersection.y, 5)


class TestRect(TestCase):
    def test_link_from_top(self):
        """Test linking rectangles from the top"""
        r1 = Rect(Point2D(0, 0), 10, 10)
        r2 = Rect(Point2D(0, 100), 10, 10)
        link = r1.link(r2)
        self.assertAlmostEqual(link.p0.x, 0, places=7)
        self.assertAlmostEqual(link.p0.y, 5, places=7)
        self.assertAlmostEqual(link.p1.x, 0, places=7)
        self.assertAlmostEqual(link.p1.y, 95, places=7)
        self.assertAlmostEqual(link.length(), 90, places=7)

    def test_link_from_left(self):
        """Test linking rectangles from the left"""
        r1 = Rect(Point2D(0, 0), 20, 10)
        r2 = Rect(Point2D(-100, 0), 20, 10)
        link = r1.link(r2)
        self.assertAlmostEqual(link.p0.x, -10, places=7)
        self.assertAlmostEqual(link.p0.y, 0, places=7)
        self.assertAlmostEqual(link.p1.x, -90, places=7)
        self.assertAlmostEqual(link.p1.y, 0, places=7)
        self.assertAlmostEqual(link.length(), 80, places=7)

    def test_link_corner_to_corner(self):
        """Test linking rectangles corner to corner"""
        r1 = Rect(Point2D(0, 0), 10, 10)
        r2 = Rect(Point2D(100, 100), 10, 10)
        link = r1.link(r2)
        self.assertAlmostEqual(link.p0.x, 5, places=7)
        self.assertAlmostEqual(link.p0.y, 5, places=7)
        self.assertAlmostEqual(link.p1.x, 95, places=7)
        self.assertAlmostEqual(link.p1.y, 95, places=7)

    def test_link_corner_to_corner_no_squares(self):
        """Test linking non-square rectangles corner to corner"""
        x = -8
        while x < 8:
            r1 = Rect(Point2D(0, 0), 4, 2)
            r2 = Rect(Point2D(x, 4), 4, 2)
            link = r1.link(r2)
            self.assertAlmostEqual(link.p0.y, 1, places=7)
            self.assertAlmostEqual(link.p1.y, 3, places=7)
            x += 0.1

    def test_edges_returns_edges(self):
        """Test that edges() returns correct edges"""
        rect = Rect(Point2D(0, 0), 10, 10)
        edges = rect.edges()

        # Expected edges
        e1 = Point2D(-5, -5).line_to(Point2D(5, -5))
        e2 = Point2D(5, -5).line_to(Point2D(5, 5))
        e3 = Point2D(5, 5).line_to(Point2D(-5, 5))
        e4 = Point2D(-5, 5).line_to(Point2D(-5, -5))

        expected_edges = [e1, e2, e3, e4]

        self.assertEqual(len(edges), 4)

        # Check that each expected edge is found in the actual edges
        for expected_edge in expected_edges:
            found = False
            for edge in edges:
                if (
                    edge.p0.sub(expected_edge.p0).length() < 1e-3
                    and edge.p1.sub(expected_edge.p1).length() < 1e-3
                ):
                    found = True
                    break
            self.assertTrue(
                found,
                f"Expected edge from {expected_edge.p0.x},{expected_edge.p0.y} to {expected_edge.p1.x},{expected_edge.p1.y} not found",
            )



import unittest
from unittest.mock import MagicMock

from abstra_internals.entities.execution_context import Response
from abstra_internals.settings import Settings

Settings.set_root_path("/tmp")


class TestPlayerRunPageDrainsStdioMessages(unittest.TestCase):
    """Tests that _run_page in the player blueprint correctly drains
    intermediate stdio messages and handles string responses."""

    def _setup_app_and_mock(self, recv_sequence):
        """Create a Flask test client with mocked controller and connection."""
        from abstra_internals.server.apps import get_cloud_app

        controller = MagicMock()

        page = MagicMock()
        page.id = "test-page-id"
        page.path = "test-page"
        page.file = "test_page.py"
        controller.get_page_stage_by_path = MagicMock(return_value=page)

        mock_conn = MagicMock()
        mock_conn.recv = MagicMock(side_effect=recv_sequence)
        mock_conn.poll = MagicMock(return_value=True)
        controller.repositories.producer.enqueue = MagicMock(return_value=mock_conn)

        app = get_cloud_app(controller)
        client = app.test_client()
        return client, mock_conn

    def test_normal_dict_response(self):
        """Page runner returns correct HTML when response is a dict."""
        recv_sequence = [
            '{"type":"execution:started"}',
            {
                "headers": {"Content-Type": "text/html"},
                "status": 200,
                "body": "<h1>Hello</h1>",
            },
        ]
        client, mock_conn = self._setup_app_and_mock(recv_sequence)

        res = client.get("/_page/test-page")

        self.assertEqual(res.status_code, 200)
        self.assertIn(b"<h1>Hello</h1>", res.data)
        mock_conn.close.assert_called_once()

    def test_string_response_is_deserialized(self):
        """When the response arrives as a JSON string, it is deserialized correctly."""
        recv_sequence = [
            '{"type":"execution:started"}',
            '{"headers": {"Content-Type": "text/html"}, "status": 200, "body": "<p>String</p>"}',
        ]
        client, mock_conn = self._setup_app_and_mock(recv_sequence)

        res = client.get("/_page/test-page")

        self.assertEqual(res.status_code, 200)
        self.assertIn(b"<p>String</p>", res.data)

    def test_drains_stdio_batch_before_response(self):
        """stdio_batch messages are skipped until the actual Response is found."""
        recv_sequence = [
            '{"type":"execution:started"}',
            {
                "type": "stdio_batch",
                "payload": [{"type": "stdout", "log": "debug output"}],
            },
            {"headers": {}, "status": 200, "body": "<h1>Page</h1>"},
        ]
        client, mock_conn = self._setup_app_and_mock(recv_sequence)

        res = client.get("/_page/test-page")

        self.assertEqual(res.status_code, 200)
        self.assertIn(b"<h1>Page</h1>", res.data)

    def test_drains_multiple_stdio_messages(self):
        """Multiple stdio and stdio_batch messages are all drained."""
        recv_sequence = [
            '{"type":"execution:started"}',
            {"type": "stdio_batch", "payload": [{"type": "stdout", "log": "line1"}]},
            {"type": "stdio", "payload": {"type": "stderr", "log": "warn"}},
            {"type": "stdio_batch", "payload": [{"type": "stdout", "log": "line2"}]},
            {
                "headers": {"Content-Type": "application/json"},
                "status": 200,
                "body": '{"result": 42}',
            },
        ]
        client, mock_conn = self._setup_app_and_mock(recv_sequence)

        res = client.post("/_page/test-page")

        self.assertEqual(res.status_code, 200)
        self.assertIn(b'{"result": 42}', res.data)

    def test_string_execution_ended_returns_500(self):
        """When execution:ended arrives as string (no Response sent),
        it should return 500 — the worker finished without sending a page response."""
        recv_sequence = [
            '{"type":"execution:started"}',
            '{"type":"execution:ended","data":{"exitStatus":"EXCEPTION"}}',
        ]
        client, mock_conn = self._setup_app_and_mock(recv_sequence)

        res = client.get("/_page/test-page")

        self.assertEqual(res.status_code, 500)

    def test_response_object_used_directly(self):
        """When connection returns a Response instance, it's used directly."""
        recv_sequence = [
            '{"type":"execution:started"}',
            Response(headers={"Content-Type": "text/plain"}, status=200, body="direct"),
        ]
        client, mock_conn = self._setup_app_and_mock(recv_sequence)

        res = client.get("/_page/test-page")

        self.assertEqual(res.status_code, 200)
        self.assertIn(b"direct", res.data)

    def test_stream_response_still_works(self):
        """Streaming responses (generator functions) still work after the fix."""
        recv_sequence = [
            '{"type":"execution:started"}',
            {
                "__page_stream__": "start",
                "status": 200,
                "headers": {"Content-Type": "application/x-ndjson"},
            },
        ]
        # After stream start, the generate() function will call recv() for chunks
        client, mock_conn = self._setup_app_and_mock(recv_sequence)

        # Add chunk responses for the streaming part
        mock_conn.recv.side_effect = [
            recv_sequence[0],  # execution:started
            recv_sequence[1],  # stream start
            {"__page_stream__": "chunk", "data": "hello"},
            {"__page_stream__": "end"},
        ]

        res = client.get("/_page/test-page")

        self.assertEqual(res.status_code, 200)

    def test_stdio_before_stream_start(self):
        """stdio messages before a stream start are drained correctly."""
        recv_sequence = [
            '{"type":"execution:started"}',
            {"type": "stdio_batch", "payload": []},
            {
                "__page_stream__": "start",
                "status": 200,
                "headers": {"Content-Type": "application/x-ndjson"},
            },
        ]
        client, mock_conn = self._setup_app_and_mock(recv_sequence)

        mock_conn.recv.side_effect = [
            recv_sequence[0],
            recv_sequence[1],
            recv_sequence[2],
            {"__page_stream__": "end"},
        ]

        res = client.get("/_page/test-page")

        self.assertEqual(res.status_code, 200)



import shutil
from pathlib import Path
from typing import Union
from uuid import uuid4

from abstra_internals.constants import (
    get_persistent_dir,
    get_project_url,
    get_public_dir,
)
from abstra_internals.settings import Settings


def create_public_link(path: Union[str, Path]) -> str:
    """
    Create a public link for a file by copying it to the public directory

    Args:
        path (Union[str, pathlib.Path]): Path to the file

    Returns:
        str: Public link to the file
    """

    if isinstance(path, str):
        path = Settings.root_path / path

    if not path.exists():
        raise Exception("path does not exist")

    new_path = f"{uuid4()}/{path.name}"
    public_path = get_public_dir() / new_path
    public_path.parent.mkdir(parents=True, exist_ok=True)

    shutil.copy(path, public_path)
    return f"{get_project_url()}/_public/{new_path}"


__all__ = [
    "get_public_dir",
    "get_persistent_dir",
    "create_public_link",
]
